class BotException(Exception):
    """
    bot 异常基类
    """

    def __init__(self, msg: str):
        super().__init__(self, msg)
        self.err = msg

    def __str__(self):
        return self.err


class PluginLoadError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class PluginBuildError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotConnectFailed(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotHookError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class EventHandlerError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotCheckerError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotMatcherError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class ShareObjError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class PluginSignalError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotActionError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotSessionError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotBaseUtilsError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class ArgParseError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class ArgFormatInitError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class ArgFormatFailed(BotException):
    """
    专指格式化器格式化失败
    """

    def __init__(self, msg: str):
        super().__init__(msg)


class ArgVerifyFailed(BotException):
    """
    格式化时参数验证不通过
    """

    def __init__(self, msg: str = ""):
        super().__init__(msg)


class ArgLackError(BotException):
    """
    格式化时参数缺失
    """

    def __init__(self, msg: str = ""):
        super().__init__(msg)


class DirectRetSignal(BotException):
    """
    嵌套函数中快速 return 的信号，上游捕获后无视即可
    """

    def __init__(self, msg: str = ""):
        super().__init__(msg)


class SessionHupTimeout(BotException):
    """
    等待 session 被唤醒时的超时异常
    """

    def __init__(self, msg: str = ""):
        super().__init__(msg)



import inspect
from asyncio import iscoroutinefunction
from pathlib import Path

from ..models.handler import EventHandler, EventHandlerArgs
from ..types.core import IActionResponder
from ..types.exceptions import *
from ..types.models import (
    HookRunnerArgs,
    SessionRule,
    ShareCbArgs,
    ShareObjArgs,
    SignalHandlerArgs,
)
from ..types.typing import *
from ..types.utils import BotChecker, BotMatcher, BotParser, Logger, PrefixLogger
from ..utils.checker import (
    AtChecker,
    FriendReqChecker,
    GroupReqChecker,
    NoticeTypeChecker,
)
from ..utils.matcher import ContainMatch, EndMatch, FullMatch, RegexMatch, StartMatch
from .bot import BotHookBus, HookRunner, PluginProxy
from .handler import (
    MetaEventHandler,
    MsgEventHandler,
    NoticeEventHandler,
    ReqEventHandler,
)
from .ipc import PluginBus, PluginSignalHandler, PluginStore
from .session import BotSessionManager


class Plugin:
    """
    bot 插件基类。所有自定义插件必须继承该类实现。
    """

    # 标记了该类有哪些实例属性需要被共享。每个元素是一个共享对象构造参数元组
    __share__: List[ShareObjArgs] = []
    # 插件 id 和 version 标记
    __id__: str = None
    __version__: str = "1.0.0"
    # 插件类所在的文件路径，PosicPath 对象
    ROOT: Path
    # 被二次包装的全局日志器
    LOGGER: PrefixLogger

    def __init__(self) -> None:
        self.__handlers: List[EventHandler]
        self.__proxy: PluginProxy

    def __build(
        self, root_path: Path, logger: Logger, responder: IActionResponder
    ) -> None:
        """
        初始化当前插件
        """
        if self.__class__.__id__ is None:
            self.__class__.__id__ = self.__class__.__name__
        for idx, val in enumerate(self.__class__.__share__):
            if isinstance(val, str):
                self.__class__.__share__[idx] = val, self.__class__.__name__, val
        self.__handlers = []

        self.__class__.ROOT = root_path
        self.__class__.LOGGER = PrefixLogger(logger, self.__class__.__id__)

        attrs_map = {
            k: v for k, v in inspect.getmembers(self) if not k.startswith("__")
        }
        for val in self.__class__.__share__:
            property, namespace, id = val
            if property not in attrs_map.keys() and property is not None:
                raise PluginBuildError(
                    f"插件 {self.__class__.__name__} 尝试共享一个不存在的属性 {property}"
                )
            PluginStore._create_so(property, namespace, id, self)

        members = inspect.getmembers(self)
        for attr_name, val in members:
            if isinstance(val, EventHandlerArgs):
                executor, handler_class, params = val
                if not iscoroutinefunction(executor):
                    raise PluginBuildError(f"事件处理器 {executor.__name__} 必须为异步方法")
                overtime_cb_maker, conflict_cb_maker = params[-1], params[-2]
                if overtime_cb_maker and not callable(overtime_cb_maker):
                    raise PluginBuildError(
                        f"超时回调方法 {overtime_cb_maker.__name__} 必须为可调用对象"
                    )
                if conflict_cb_maker and not callable(conflict_cb_maker):
                    raise PluginBuildError(
                        f"冲突回调方法 {conflict_cb_maker.__name__} 必须为可调用对象"
                    )
                handler = handler_class(executor, self, responder, logger, *params)
                self.__handlers.append(handler)
                BotSessionManager.register(handler)

            elif isinstance(val, HookRunnerArgs):
                hook_func, type = val
                if not iscoroutinefunction(hook_func):
                    raise PluginBuildError(f"hook 方法 {hook_func.__name__} 必须为异步函数")
                runner = HookRunner(type, hook_func, self)
                BotHookBus._register(type, runner)

            elif isinstance(val, ShareCbArgs):
                namespace, id, cb = val
                if not iscoroutinefunction(cb):
                    raise PluginBuildError(
                        f"{namespace} 命名空间中，id 为 {id} 的共享对象，它的回调 {cb.__name__} 必须为异步函数"
                    )
                PluginStore._bind_cb(namespace, id, cb, self)

            elif isinstance(val, SignalHandlerArgs):
                func, namespace, signal = val
                if not iscoroutinefunction(func):
                    raise PluginBuildError(f"信号处理方法 {func.__name__} 必须为异步函数")
                handler = PluginSignalHandler(namespace, signal, func, self)
                PluginBus._register(namespace, signal, handler)

        self.__proxy = PluginProxy(self)

    @classmethod
    def on_message(
        cls,
        matcher: BotMatcher = None,
        parser: BotParser = None,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: int = None,
        block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[None], Coroutine] = None,
        overtime_cb: Callable[[None], Coroutine] = None,
    ) -> Callable:
        """
        使用该装饰器，将方法标记为消息事件执行器
        """

        def make_args(executor: AsyncFunc[None]) -> EventHandlerArgs:
            return EventHandlerArgs(
                executor=executor,
                type=MsgEventHandler,
                params=[
                    matcher,
                    parser,
                    checker,
                    priority,
                    timeout,
                    block,
                    temp,
                    session_rule,
                    session_hold,
                    direct_rouse,
                    conflict_wait,
                    conflict_cb,
                    overtime_cb,
                ],
            )

        return make_args

    @classmethod
    def on_every_message(
        cls,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: int = None,
        block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[None], Coroutine] = None,
        overtime_cb: Callable[[None], Coroutine] = None,
    ) -> Callable:
        """
        使用该装饰器，将方法标记为任意消息事件执行器。
        任何消息经过校验后，不进行匹配和解析即可触发处理方法
        """

        def make_args(executor: AsyncFunc[None]) -> EventHandlerArgs:
            return EventHandlerArgs(
                executor=executor,
                type=MsgEventHandler,
                params=[
                    None,
                    None,
                    checker,
                    priority,
                    timeout,
                    block,
                    temp,
                    session_rule,
                    session_hold,
                    direct_rouse,
                    conflict_wait,
                    conflict_cb,
                    overtime_cb,
                ],
            )

        return make_args

    @classmethod
    def on_at_qq(
        cls,
        qid: int = None,
        matcher: BotMatcher = None,
        parser: BotParser = None,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: int = None,
        block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[None], Coroutine] = None,
        overtime_cb: Callable[[None], Coroutine] = None,
    ) -> Callable:
        """
        使用该装饰器，将方法标记为艾特消息匹配的消息事件执行器。
        必须首先是来自指定 qid 的艾特消息，才能被进一步处理
        """
        at_checker = AtChecker(qid)
        if checker is not None:
            wrapped_checker = at_checker & checker
        else:
            wrapped_checker = at_checker

        def make_args(executor: AsyncFunc[None]) -> EventHandlerArgs:
            return EventHandlerArgs(
                executor=executor,
                type=MsgEventHandler,
                params=[
                    matcher,
                    parser,
                    wrapped_checker,
                    priority,
                    timeout,
                    block,
                    temp,
                    session_rule,
                    session_hold,
                    direct_rouse,
                    conflict_wait,
                    conflict_cb,
                    overtime_cb,
                ],
            )

        return make_args

    @classmethod
    def on_start_match(
        cls,
        target: str,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: int = None,
        block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[None], Coroutine] = None,
        overtime_cb: Callable[[None], Coroutine] = None,
    ) -> Callable:
        """
        使用该装饰器，将方法标记为字符串起始匹配的消息事件执行器。
        必须首先含有以 target 起始的文本，才能被进一步处理
        """
        start_matcher = StartMatch(target)

        def make_args(executor: AsyncFunc[None]) -> EventHandlerArgs:
            return EventHandlerArgs(
                executor=executor,
                type=MsgEventHandler,
                params=[
                    start_matcher,
                    None,
                    checker,
                    priority,
                    timeout,
                    block,
                    temp,
                    session_rule,
                    session_hold,
                    direct_rouse,
                    conflict_wait,
                    conflict_cb,
                    overtime_cb,
                ],
            )

        return make_args

    @classmethod
    def on_contain_match(
        cls,
        target: str,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: int = None,
        block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[None], Coroutine] = None,
        overtime_cb: Callable[[None], Coroutine] = None,
    ) -> Callable:
        """
        使用该装饰器，将方法标记为字符串包含匹配的消息事件执行器。
        文本必须首先包含 target，才能被进一步处理
        """
        contain_matcher = ContainMatch(target)

        def make_args(executor: AsyncFunc[None]) -> EventHandlerArgs:
            return EventHandlerArgs(
                executor=executor,
                type=MsgEventHandler,
                params=[
                    contain_matcher,
                    None,
                    checker,
                    priority,
                    timeout,
                    block,
                    temp,
                    session_rule,
                    session_hold,
                    direct_rouse,
                    conflict_wait,
                    conflict_cb,
                    overtime_cb,
                ],
            )

        return make_args

    @classmethod
    def on_full_match(
        cls,
        target: str,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: int = None,
        block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[None], Coroutine] = None,
        overtime_cb: Callable[[None], Coroutine] = None,
    ) -> Callable:
        """
        使用该装饰器，将方法标记为字符串相等匹配的消息事件执行器。
        文本必须首先与 target 内容完全一致，才能被进一步处理
        """
        full_matcher = FullMatch(target)

        def make_args(executor: AsyncFunc[None]) -> EventHandlerArgs:
            return EventHandlerArgs(
                executor=executor,
                type=MsgEventHandler,
                params=[
                    full_matcher,
                    None,
                    checker,
                    priority,
                    timeout,
                    block,
                    temp,
                    session_rule,
                    session_hold,
                    direct_rouse,
                    conflict_wait,
                    conflict_cb,
                    overtime_cb,
                ],
            )

        return make_args

    @classmethod
    def on_end_match(
        cls,
        target: str,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: int = None,
        block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[None], Coroutine] = None,
        overtime_cb: Callable[[None], Coroutine] = None,
    ) -> Callable:
        """
        使用该装饰器，将方法标记为字符串结尾匹配的消息事件执行器。
        文本必须首先以 target 结尾，才能被进一步处理
        """
        end_matcher = EndMatch(target)

        def make_args(executor: AsyncFunc[None]) -> EventHandlerArgs:
            return EventHandlerArgs(
                executor=executor,
                type=MsgEventHandler,
                params=[
                    end_matcher,
                    None,
                    checker,
                    priority,
                    timeout,
                    block,
                    temp,
                    session_rule,
                    session_hold,
                    direct_rouse,
                    conflict_wait,
                    conflict_cb,
                    overtime_cb,
                ],
            )

        return make_args

    @classmethod
    def on_regex_match(
        cls,
        target: str,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: int = None,
        block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[None], Coroutine] = None,
        overtime_cb: Callable[[None], Coroutine] = None,
    ) -> Callable:
        """
        使用该装饰器，将方法标记为字符串正则匹配的消息事件执行器。
        文本必须包含指定的正则内容，才能被进一步处理
        """
        regex_matcher = RegexMatch(target)

        def make_args(executor: AsyncFunc[None]) -> EventHandlerArgs:
            return EventHandlerArgs(
                executor=executor,
                type=MsgEventHandler,
                params=[
                    regex_matcher,
                    None,
                    checker,
                    priority,
                    timeout,
                    block,
                    temp,
                    session_rule,
                    session_hold,
                    direct_rouse,
                    conflict_wait,
                    conflict_cb,
                    overtime_cb,
                ],
            )

        return make_args

    @classmethod
    def on_request(
        cls,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: int = None,
        block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[None], Coroutine] = None,
        overtime_cb: Callable[[None], Coroutine] = None,
    ) -> Callable:
        """
        使用该装饰器，将方法标记为请求事件执行器
        """

        def make_args(executor: AsyncFunc[None]) -> EventHandlerArgs:
            return EventHandlerArgs(
                executor=executor,
                type=ReqEventHandler,
                params=[
                    checker,
                    priority,
                    timeout,
                    block,
                    temp,
                    session_rule,
                    session_hold,
                    direct_rouse,
                    conflict_wait,
                    conflict_cb,
                    overtime_cb,
                ],
            )

        return make_args

    @classmethod
    def on_friend_request(
        cls,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: int = None,
        block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[None], Coroutine] = None,
        overtime_cb: Callable[[None], Coroutine] = None,
    ) -> Callable:
        """
        使用该装饰器，将方法标记为私聊请求事件执行器
        """
        friend_checker = FriendReqChecker()
        if checker is not None:
            wrapped_checker = friend_checker & checker
        else:
            wrapped_checker = friend_checker

        def make_args(executor: AsyncFunc[None]) -> EventHandlerArgs:
            return EventHandlerArgs(
                executor=executor,
                type=ReqEventHandler,
                params=[
                    wrapped_checker,
                    priority,
                    timeout,
                    block,
                    temp,
                    session_rule,
                    session_hold,
                    direct_rouse,
                    conflict_wait,
                    conflict_cb,
                    overtime_cb,
                ],
            )

        return make_args

    @classmethod
    def on_group_request(
        cls,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: int = None,
        block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[None], Coroutine] = None,
        overtime_cb: Callable[[None], Coroutine] = None,
    ) -> Callable:
        """
        使用该装饰器，将方法标记为群请求事件执行器
        """
        group_checker = GroupReqChecker()
        if checker is not None:
            wrapped_checker = group_checker & checker
        else:
            wrapped_checker = group_checker

        def make_args(executor: AsyncFunc[None]) -> EventHandlerArgs:
            return EventHandlerArgs(
                executor=executor,
                type=ReqEventHandler,
                params=[
                    wrapped_checker,
                    priority,
                    timeout,
                    block,
                    temp,
                    session_rule,
                    session_hold,
                    direct_rouse,
                    conflict_wait,
                    conflict_cb,
                    overtime_cb,
                ],
            )

        return make_args

    @classmethod
    def on_notice(
        cls,
        type: str = Literal[
            "group_upload",
            "group_admin",
            "group_decrease",
            "group_increase",
            "group_ban",
            "friend_add",
            "group_recall",
            "friend_recall",
            "group_card",
            "offline_file",
            "client_status",
            "essence",
            "notify",
            "honor",
            "poke",
            "lucky_king",
            "title",
            "ALL",
        ],
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: int = None,
        block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[None], Coroutine] = None,
        overtime_cb: Callable[[None], Coroutine] = None,
    ) -> Callable:
        """
        使用该装饰器，将方法标记为通知事件执行器
        """
        type_checker = NoticeTypeChecker(type)
        if checker is not None:
            wrapped_checker = type_checker & checker
        else:
            wrapped_checker = type_checker

        def make_args(executor: AsyncFunc[None]) -> EventHandlerArgs:
            return EventHandlerArgs(
                executor=executor,
                type=NoticeEventHandler,
                params=[
                    wrapped_checker,
                    priority,
                    timeout,
                    block,
                    temp,
                    session_rule,
                    session_hold,
                    direct_rouse,
                    conflict_wait,
                    conflict_cb,
                    overtime_cb,
                ],
            )

        return make_args

    @classmethod
    def on_meta_event(
        cls,
        checker: BotChecker = None,
        priority: PriorityLevel = PriorityLevel.MEAN,
        timeout: int = None,
        block: bool = False,
        temp: bool = False,
        session_rule: SessionRule = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Callable[[None], Coroutine] = None,
        overtime_cb: Callable[[None], Coroutine] = None,
    ) -> Callable:
        """
        使用该装饰器，将方法标记为元事件执行器
        """

        def make_args(executor: AsyncFunc[None]) -> EventHandlerArgs:
            return EventHandlerArgs(
                executor=executor,
                type=MetaEventHandler,
                params=[
                    checker,
                    priority,
                    timeout,
                    block,
                    temp,
                    session_rule,
                    session_hold,
                    direct_rouse,
                    conflict_wait,
                    conflict_cb,
                    overtime_cb,
                ],
            )

        return make_args



import json
import sys

sys.path.append(r"E:\projects\Python\git-proj\melobot")
from melobot import to_cq_arr

d = to_cq_arr(
    "[CQ:image,id=123456,id2=123456.789a]asdjfa;jfajfa[CQ:image,id=asdjf;a;f,id2=asdjf;a;fa;jfadsf];sdjf;asfj[CQ:image,id=asdjf;a;f,id2=asdjf;a;fa;jfadsf]"
)
print(json.dumps(d, ensure_ascii=False, indent=2))



import asyncio as aio
import os
import time
import traceback
from asyncio import iscoroutinefunction
from functools import partial
from types import MethodType

from ..meta import MODULE_MODE_FLAG, MODULE_MODE_SET
from ..types.core import IActionResponder
from ..types.exceptions import *
from ..types.models import BotLife, HookRunnerArgs
from ..types.typing import *
from ..types.utils import Logger
from ..utils.config import BotConfig
from .ipc import PluginStore, ShareObject
from .session import SESSION_LOCAL, BotSessionManager


class HookRunner:
    """
    bot 生命周期 hook 运行器
    """

    def __init__(self, type: str, func: Callable, plugin: Union[object, None]) -> None:
        self._func = func
        self._plugin = plugin
        # 对应：绑定的 func 是插件类的实例方法
        if plugin:
            self.cb = partial(self._func, plugin)
        # 对应：绑定的 func 是普通函数
        else:
            self.cb = self._func
        self.type = type


class BotHookBus:
    """
    bot 生命周期 hook 总线
    """

    __store__: Dict[BotLife, List[HookRunner]] = {
        v: [] for k, v in BotLife.__members__.items()
    }
    __logger: Logger
    __responder: IActionResponder

    @classmethod
    def _bind(cls, logger: Logger, responder: IActionResponder) -> None:
        """
        初始化该类，绑定全局日志器和行为响应器
        """
        cls.__logger = logger
        cls.__responder = responder

    @classmethod
    def _register(cls, hook_type: BotLife, runner: HookRunner) -> None:
        """
        注册一个生命周期运行器。由 plugin build 过程调用
        """
        if hook_type not in cls.__store__.keys():
            raise BotHookError(
                f"尝试添加一个生命周期 hook，但是其指定的类型 {hook_type} 不存在"
            )
        cls.__store__[hook_type].append(runner)

    @classmethod
    def on(cls, hook_type: BotLife, callback: Callable) -> None:
        """
        动态注册 hook 方法
        """
        if isinstance(callback, HookRunner):
            raise BotHookError("已注册的生命周期 hook 方法不能再注册")
        if not iscoroutinefunction(callback):
            raise BotHookError(f"生命周期 hook 方法 {callback.__name__} 必须为异步函数")
        if (
            isinstance(callback, partial) and isinstance(callback.func, MethodType)
        ) or isinstance(callback, MethodType):
            raise BotHookError("callback 应该是 function，而不是 bound method")
        runner = HookRunner(hook_type, callback, plugin=None)
        cls._register(hook_type, runner)

    @classmethod
    async def _run_on_ctx(cls, runner: HookRunner, *args, **kwargs) -> None:
        session = BotSessionManager.make_empty(cls.__responder)
        token = SESSION_LOCAL._add_ctx(session)
        try:
            await runner.cb(*args, **kwargs)
        except Exception as e:
            e_name = e.__class__.__name__
            func_name = runner._func.__qualname__
            pre_str = (
                "插件" + runner._plugin.__class__.__id__
                if runner._plugin
                else "动态注册的"
            )
            cls.__logger.error(
                f"{pre_str} hook 方法 {func_name} 发生异常：[{e_name}] {e}"
            )
            cls.__logger.debug(f"生命周期 hook 方法的 args: {args} kwargs：{kwargs}")
            cls.__logger.debug("异常回溯栈：\n" + traceback.format_exc().strip("\n"))
        finally:
            SESSION_LOCAL._del_ctx(token)

    @classmethod
    async def emit(
        cls, hook_type: BotLife, *args, wait: bool = False, **kwargs
    ) -> None:
        """
        触发一个生命周期信号。如果指定 wait 为 True，则会等待所有生命周期 hook 方法完成
        """
        if not wait:
            for runner in cls.__store__[hook_type]:
                aio.create_task(cls._run_on_ctx(runner, *args, **kwargs))
        else:
            for runner in cls.__store__[hook_type]:
                await cls._run_on_ctx(runner, *args, **kwargs)


class PluginProxy:
    """
    bot 插件代理类。供外部使用
    """

    def __init__(self, plugin: object) -> None:
        self.id = plugin.__class__.__id__
        self.version = plugin.__class__.__version__
        self.root_path = plugin.__class__.ROOT
        self.share: Dict[str, ShareObject] = {}
        for property, namespace, id in plugin.__class__.__share__:
            self.share[id] = PluginStore.get(namespace, id)


class BotProxy:
    """
    外部使用的 Bot 对象
    """

    def __init__(self) -> None:
        self.__origin__: object

    def _bind(self, bot_origin: object) -> None:
        self.__origin__ = bot_origin

    @property
    def loop(self) -> aio.AbstractEventLoop:
        return self.__origin__.loop

    @property
    def logger(self) -> Logger:
        """
        bot 全局日志器
        """
        return self.__origin__.logger

    @property
    def config(self) -> BotConfig:
        """
        bot 全局配置
        """
        return self.__origin__.config

    @property
    def plugins(self) -> Dict[str, PluginProxy]:
        """
        获取 bot 当前所有插件信息
        """
        return {
            id: plugin._Plugin__proxy for id, plugin in self.__origin__.plugins.items()
        }

    @property
    def is_activate(self) -> bool:
        return not self.__origin__.slack

    def slack(self) -> None:
        """
        使 bot 不再发送 action。但 ACTION_PRESEND 钩子依然会触发
        """
        self.__origin__.slack = True

    def activate(self) -> None:
        """
        使 bot 可以发送 action
        """
        self.__origin__.slack = False

    async def close(self) -> None:
        """
        关闭 bot
        """
        await self.__origin__.close()

    def is_module_run(self) -> bool:
        """
        检查是否是模块运行模式
        """
        return os.environ.get(MODULE_MODE_FLAG) == MODULE_MODE_SET

    async def restart(self) -> None:
        """
        重启 bot。只可在模块运行模式下使用
        """
        await self.__origin__.restart()

    def call_later(self, callback: partial, delay: float):
        """
        在指定的 delay 后调度一个 callback 执行。注意这个 callback 应该是同步方法
        """
        return self.loop.call_later(delay, callback)

    def call_at(self, callback: partial, timestamp: float):
        """
        在指定的时间戳后调度一个 callback 执行。注意这个 callback 应该是同步方法

        当 timestamp <= 当前时刻将立即执行
        """
        if timestamp <= time.time():
            return self.loop.call_soon(callback)
        else:
            return self.loop.call_later(timestamp - time.time(), callback)

    def async_later(
        self, callback: Coroutine[Any, Any, Any], delay: float
    ) -> aio.Future:
        """
        在指定的 delay 后调度一个 callback 执行。注意这个 callback 应该是协程。
        返回一个 future 对象，可用于等待结果
        """

        async def async_cb(fut: aio.Future) -> Any:
            await aio.sleep(delay)
            res = await callback
            fut.set_result(res)

        fut = aio.Future()
        aio.create_task(async_cb(fut))
        return fut

    def async_at(
        self, callback: Coroutine[Any, Any, Any], timestamp: float
    ) -> aio.Future:
        """
        在指定的 timestamp 调度一个 callback 执行。注意这个 callback 应该是协程。
        返回一个 future 对象，可用于等待结果

        当 timestamp <= 当前时刻将立即执行
        """
        if timestamp <= time.time():
            return self.async_later(callback, 0)
        else:
            return self.async_later(callback, timestamp - time.time())

    def async_interval(
        self, cb_maker: Callable[[None], Coroutine[Any, Any, None]], interval: float
    ) -> aio.Task:
        """
        以指定的 interval 调度一个 callback 执行。cb_maker 是用于产生 callback 的协程产生器。
        返回一个 task 对象用于取消 callback
        """

        async def interval_cb():
            while True:
                await aio.sleep(interval)
                await cb_maker()

        t = aio.create_task(interval_cb())
        return t

    @classmethod
    def on(cls, hook_type: BotLife, callback: Callable = None):
        """
        用作装饰器，不传入 callback 参数，即可注册一个 bot 生命周期 hook。

        也可直接调用此方法，传入 callback 参数来注册一个 hook。
        callback 可以是类实例方法，也可以不是。callback 如果是类实例方法，请自行包裹为一个 partial 函数。

        例如你的插件类是：`A`，而你需要传递一个类实例方法：`A.xxx`，
        那么你的 callback 参数就应该是：`functools.partial(A.xxx, self)`
        """

        def make_args(func: AsyncFunc[None]) -> HookRunnerArgs:
            return HookRunnerArgs(func=func, type=hook_type)

        if callback is None:
            return make_args
        else:
            BotHookBus.on(hook_type, callback)


BOT_PROXY = BotProxy()



"""
MeloBot 是插件化管理、基于异步会话机制的 qbot 开发框架。

项目源码和文档，请参考：https://github.com/aicorein/melobot
"""

from .core.init import MeloBot
from .meta import META_INFO, MetaInfo
from .models import (
    AttrSessionRule,
    BotSession,
    MetaEvent,
    MsgEvent,
    NoticeEvent,
    Plugin,
    PluginBus,
    PluginStore,
    RequestEvent,
    RespEvent,
    RWController,
    bot,
    event,
    finish,
    get_twin_event,
    msg_text,
    send,
    send_hup,
    send_reply,
    session,
    to_cq_arr,
)
from .types import (
    BotChecker,
    BotEvent,
    BotLife,
    BotMatcher,
    PriorityLevel,
    SessionRule,
    ShareObjArgs,
    User,
)
from .utils import (
    AlwaysMatch,
    ArgFormatter,
    CmdParser,
    CmdParserGen,
    ContainMatch,
    EndMatch,
    FullMatch,
    GroupMsgLvl,
    MsgCheckerGen,
    MsgLvlChecker,
    PrivateMsgLvl,
    RegexMatch,
    StartMatch,
    cooldown,
    get_cq_text,
    get_id,
    lock,
    semaphore,
    this_dir,
    to_async,
    to_coro,
)

session: BotSession


def get_metainfo() -> MetaInfo:
    """
    获取 melobot 核心元信息，外部更改无效
    """
    return MetaInfo()


__version__ = META_INFO.VER
__all__ = (
    "MeloBot",
    "get_metainfo",
    "AttrSessionRule",
    "session",
    "to_cq_arr",
    "MetaEvent",
    "MsgEvent",
    "NoticeEvent",
    "Plugin",
    "PluginBus",
    "PluginStore",
    "RequestEvent",
    "RespEvent",
    "RWController",
    "bot",
    "event",
    "msg_text",
    "finish",
    "get_twin_event",
    "send",
    "send_hup",
    "send_reply",
    "BotChecker",
    "BotEvent",
    "BotLife",
    "BotMatcher",
    "PriorityLevel",
    "SessionRule",
    "ShareObjArgs",
    "User",
    "AlwaysMatch",
    "ArgFormatter",
    "CmdParser",
    "CmdParserGen",
    "ContainMatch",
    "EndMatch",
    "FullMatch",
    "GroupMsgLvl",
    "MsgCheckerGen",
    "MsgLvlChecker",
    "PrivateMsgLvl",
    "RegexMatch",
    "StartMatch",
    "get_cq_text",
    "cooldown",
    "get_id",
    "lock",
    "semaphore",
    "this_dir",
    "to_async",
    "to_coro",
)



from enum import Enum
from types import ModuleType
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Coroutine,
    Dict,
    List,
    Literal,
    NamedTuple,
    Optional,
    OrderedDict,
    Set,
    Tuple,
    Type,
    TypedDict,
    TypeVar,
    Union,
)

__all__ = (
    "Union",
    "List",
    "Set",
    "Dict",
    "Literal",
    "Coroutine",
    "Callable",
    "Tuple",
    "Any",
    "OrderedDict",
    "Optional",
    "ModuleType",
    "AsyncIterator",
    "NamedTuple",
    "AsyncFunc",
    "Type",
    "Enum",
    "CQMsgDict",
    "MsgNodeDict",
    "User",
    "PriorityLevel",
    "ParseArgs",
    "Null",
)


class CQMsgDict(TypedDict):
    """
    cq 消息 dict
    """

    type: str
    data: Dict[str, Union[float, int, str]]


class MsgNodeDict(TypedDict):
    """
    消息节点 dict
    """

    type: str
    data: Dict[str, Union[float, int, str]]


class ParseArgs:
    """
    命令参数类
    """

    def __init__(self, values: List[Any]) -> None:
        self.vals = values
        self.formatted = False


class User(int, Enum):
    """
    用户权限等级枚举
    """

    OWNER = 100
    SU = 90
    WHITE = 80
    USER = 70
    BLACK = -1


class PriorityLevel(int, Enum):
    """
    优先级枚举。方便进行优先级比较，有 MIN, MAX, MEAN 三个枚举值
    """

    MIN = -100
    MAX = 100
    MEAN = (MAX + MIN) // 2


T = TypeVar("T")
AsyncFunc = Callable[..., Coroutine[Any, Any, T]]


class Null:
    pass



import asyncio
import sys

sys.path.append(r"E:\projects\Python\git-proj\melobot")
from dev.plugins.public_utils import FormData, async_http, get_headers


async def main():
    with open(r"2939fc304d7d88447f460b76aa9d7199.jpeg", "rb") as fp:
        image = fp.read()
    url = "https://api.trace.moe/search?cutBorders"
    data = FormData()
    data.add_field("image", image)
    async with async_http(url, method="post", headers=get_headers(), data=data) as resp:
        res = await resp.json()
        print(res)


asyncio.run(main())



from abc import ABC, abstractmethod
from collections.abc import Mapping
from logging import Logger
from types import TracebackType
from typing import TypeAlias

from ..models.event import BotEvent
from .exceptions import *
from .typing import *

_SysExcInfoType: TypeAlias = Union[
    tuple[type[BaseException], BaseException, Optional[TracebackType]],
    tuple[None, None, None],
]
_ExcInfoType: TypeAlias = Union[None, bool, _SysExcInfoType, BaseException]


class PrefixLogger:
    """
    二次包装的日志器
    """

    def __init__(self, ref: Logger, prefix: str) -> None:
        self._logger = ref
        self._prefix = prefix

    def _add_prefix(self, s: str) -> str:
        return f"[{self._prefix}] {s}"

    def info(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Union[Mapping[str, object], None] = None,
    ) -> None:
        msg = self._add_prefix(msg)
        return self._logger.info(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )

    def warn(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Union[Mapping[str, object], None] = None,
    ) -> None:
        msg = self._add_prefix(msg)
        return self._logger.warn(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )

    def warning(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Union[Mapping[str, object], None] = None,
    ) -> None:
        msg = self._add_prefix(msg)
        return self._logger.warning(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )

    def error(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Union[Mapping[str, object], None] = None,
    ) -> None:
        msg = self._add_prefix(msg)
        return self._logger.error(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )

    def debug(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Union[Mapping[str, object], None] = None,
    ) -> None:
        msg = self._add_prefix(msg)
        return self._logger.debug(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )

    def critical(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Union[Mapping[str, object], None] = None,
    ) -> None:
        msg = self._add_prefix(msg)
        return self._logger.critical(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )


class LogicMode(Enum):
    """
    逻辑模式枚举
    """

    AND = 1
    OR = 2
    NOT = 3
    XOR = 4


class BotChecker(ABC):
    def __init__(self) -> None:
        super().__init__()

    def __and__(self, other: "BotChecker") -> "WrappedChecker":
        if not isinstance(other, BotChecker):
            raise BotCheckerError(f"联合检查器定义时出现了非检查器对象，其值为：{other}")
        return WrappedChecker(LogicMode.AND, self, other)

    def __or__(self, other: "BotChecker") -> "WrappedChecker":
        if not isinstance(other, BotChecker):
            raise BotCheckerError(f"联合检查器定义时出现了非检查器对象，其值为：{other}")
        return WrappedChecker(LogicMode.OR, self, other)

    def __invert__(self) -> "WrappedMatcher":
        return WrappedChecker(LogicMode.NOT, self)

    def __xor__(self, other: "BotChecker") -> "WrappedChecker":
        if not isinstance(other, BotChecker):
            raise BotCheckerError(f"联合检查器定义时出现了非检查器对象，其值为：{other}")
        return WrappedChecker(LogicMode.XOR, self, other)

    @abstractmethod
    def check(self, event: BotEvent) -> bool:
        pass


class WrappedChecker(BotChecker):
    """
    按逻辑关系工作的的合并检查器，使用 AND, OR, XOR 模式时，
    需要传递两个 checker。使用 NOT 时只需要传递第一个 checker
    """

    def __init__(
        self, mode: LogicMode, checker1: BotChecker, checker2: BotChecker = None
    ) -> None:
        super().__init__()
        self.mode = mode
        self.c1, self.c2 = checker1, checker2

    def check(self, event: BotEvent) -> bool:
        if self.mode == LogicMode.AND:
            return self.c1.check(event) and self.c2.check(event)
        elif self.mode == LogicMode.OR:
            return self.c1.check(event) or self.c2.check(event)
        elif self.mode == LogicMode.NOT:
            return not self.c1.check(event)
        elif self.mode == LogicMode.XOR:
            return self.c1.check(event) ^ self.c2.check(event)


class BotMatcher(ABC):
    def __init__(self) -> None:
        super().__init__()

    def __and__(self, other: "BotMatcher") -> "WrappedMatcher":
        if not isinstance(other, BotMatcher):
            raise BotMatcherError(f"联合匹配器定义时出现了非匹配器对象，其值为：{other}")
        return WrappedMatcher(LogicMode.AND, self, other)

    def __or__(self, other: "BotMatcher") -> "WrappedMatcher":
        if not isinstance(other, BotMatcher):
            raise BotMatcherError(f"联合匹配器定义时出现了非匹配器对象，其值为：{other}")
        return WrappedMatcher(LogicMode.OR, self, other)

    def __invert__(self) -> "WrappedMatcher":
        return WrappedMatcher(LogicMode.NOT, self)

    def __xor__(self, other: "BotMatcher") -> "WrappedMatcher":
        if not isinstance(other, BotMatcher):
            raise BotMatcherError(f"联合匹配器定义时出现了非匹配器对象，其值为：{other}")
        return WrappedMatcher(LogicMode.XOR, self, other)

    @abstractmethod
    def match(self, text: str) -> bool:
        pass


class WrappedMatcher(BotMatcher):
    """
    按逻辑关系工作的的合并匹配器，使用 AND, OR, XOR 模式时，
    需要传递两个 matcher。使用 NOT 时只需要传递第一个 matcher
    """

    def __init__(
        self, mode: LogicMode, matcher1: BotMatcher, matcher2: BotMatcher = None
    ) -> None:
        super().__init__()
        self.mode = mode
        self.m1, self.m2 = matcher1, matcher2

    def match(self, text: str) -> bool:
        if self.mode == LogicMode.AND:
            return self.m1.match(text) and self.m2.match(text)
        elif self.mode == LogicMode.OR:
            return self.m1.match(text) or self.m2.match(text)
        elif self.mode == LogicMode.NOT:
            return not self.m1.match(text)
        elif self.mode == LogicMode.XOR:
            return self.m1.match(text) ^ self.m2.match(text)


class BotParser(ABC):
    """
    解析器基类。解析器一般用作从消息文本中按规则提取指定字符串或字符串组合
    """

    def __init__(self, id: Any) -> None:
        super().__init__()
        self.id = id
        self.need_format: bool = False

    @abstractmethod
    def parse(self, text: str) -> Union[Dict[str, ParseArgs], None]:
        pass

    @abstractmethod
    def test(
        self, args_group: Dict[str, ParseArgs]
    ) -> Tuple[bool, Union[str, None], Union[ParseArgs, None]]:
        pass

    @abstractmethod
    def format(self, args: ParseArgs) -> None:
        pass



import asyncio as aio
import time
from contextvars import ContextVar, Token
from copy import deepcopy
from functools import wraps

from melobot.models.event import BotEvent

from ..models.base import get_twin_event
from ..types.core import IActionResponder
from ..types.exceptions import *
from ..types.models import SessionRule
from ..types.typing import *
from .action import *
from .event import *


class BotSession:
    """
    Bot Session 类。不需要直接实例化，必须通过 BotSessionBuilder 构造。
    """

    def __init__(self, responder: IActionResponder, space_tag: object = None) -> None:
        super().__init__()
        self.store = {}
        self.timestamp = time.time()
        self.hup_times: List[float] = []
        self.events: List[
            Union[MsgEvent, RequestEvent, MetaEvent, RespEvent, NoticeEvent]
        ] = []
        self._responder = responder

        # session 是否空闲的标志，由 BotSessionManager 修改和管理
        self._free_signal = aio.Event()
        self._free_signal.set()
        # session 是否挂起的标志。二者互为孪生反状态。由 BotSessionManager 修改和管理
        # 注意 session 挂起时一定是非空闲和非过期的
        self._hup_signal, self._awake_signal = get_twin_event()
        # session 是否过期的标志，由 BotSessionManager 修改和管理
        self._expired = False
        # 用于标记该 session 属于哪个 session 空间，如果为 None 则表明是空 session 或是一次性 session
        # 其实这里如果传入 space_tag 则一定是所属 handler 的引用
        self._space_tag: Union[object, None] = space_tag

        # 所属 handler 的引用（和 space_tag 不一样，所有在 handler 中产生的 session，此属性必为非空）
        self._handler: object = None

    @property
    def event(
        self,
    ) -> Union[MsgEvent, RequestEvent, MetaEvent, RespEvent, NoticeEvent, None]:
        try:
            return next(reversed(self.events))
        except StopIteration:
            return None

    @property
    def last_hup(self) -> Union[float, None]:
        try:
            return next(reversed(self.hup_times))
        except StopIteration:
            return None

    @property
    def args(self) -> Union[Dict[str, List[Any]], List[Any], None]:
        if self._handler is not None and self.event._args_map is not None:
            args_group = self.event._args_map.get(self._handler.parser.id)
            if self._handler.parser.target is None:
                return deepcopy({name: args.vals for name, args in args_group.items()})
            for cmd_name in self._handler.parser.target:
                args = args_group.get(cmd_name)
                if args is not None:
                    return deepcopy(args.vals)
            raise BotSessionError("尝试获取的命令解析参数不存在")
        return None

    async def suspend(self, overtime: int = None) -> None:
        """
        当前 session 挂起（也就是所在方法的挂起）。直到满足同一 session_rule 的事件重新进入，
        session 所在方法便会被唤醒。但如果设置了超时时间且在唤醒前超时，则会强制唤醒 session，
        并抛出 BotHupTimeout 异常
        """
        BotSessionManager._hup(self)
        if overtime is None:
            await self._awake_signal.wait()
        elif overtime <= 0:
            raise BotSessionError("session 挂起超时时间（秒数）必须 > 0")
        else:
            await aio.wait(
                [self._awake_signal.wait(), aio.sleep(overtime)],
                return_when=aio.FIRST_COMPLETED,
            )
            if self._awake_signal.is_set():
                return
            BotSessionManager._rouse(self)
            raise SessionHupTimeout("session 挂起超时")

    def destory(self) -> None:
        """
        销毁方法。
        其他 session 调用会立即清空 session 存储、事件记录、挂起时间记录。
        如果调用 session 有 space_tag，还会从存储空间中移除该 session
        """
        if self.event is None:
            pass
        else:
            BotSessionManager._expire(self)

    def store_get(self, key: object) -> object:
        return self.store[key]

    def store_add(self, key: object, val: object) -> None:
        self.store[key] = val

    def store_update(self, store: Dict) -> None:
        self.store.update(store)

    def store_remove(self, key: object) -> None:
        self.store.pop(key)

    def store_clear(self) -> None:
        self.store.clear()

    # 不要更改这个方法下的所有 typing，否则会影响被装饰方法的 typing
    def _launch(get_action):
        """
        action 构建方法的装饰器，
        在 action 构建后进行发送，以及完成响应等待
        """

        @wraps(get_action)
        async def wrapper(self: "BotSession", *args, **kwargs):
            if self._expired:
                raise BotSessionError("session 已标记为过期，无法执行 action 操作")

            action: BotAction = await get_action(self, *args, **kwargs)
            if action.resp_id is None:
                return await self._responder.take_action(action)
            else:
                return await (await self._responder.take_action_wait(action))

        return wrapper

    """以下所有 action 方法虽然本身不是异步的，但不使用 async，返回值将没有注解"""

    @_launch
    async def custom_action(self, action: BotAction) -> Union[RespEvent, None]:
        """
        直接处理提供的 action
        """
        return action

    @_launch
    async def send(
        self,
        content: Union[str, CQMsgDict, List[CQMsgDict]],
        cq_str: bool = False,
        waitResp: bool = False,
    ) -> Union[RespEvent, None]:
        """
        在当前 session 上下文下发送消息。
        cq_str 若开启，文本中若包含 cq 字符串，将会被解释
        """
        if self.event == None:
            raise BotSessionError("空 session 需要使用 custom_send 方法替代 send")
        action = msg_action(
            content,
            self.event.is_private(),
            self.event.sender.id,
            self.event.group_id,
            waitResp,
            self.event,
        )
        if cq_str:
            action = to_cq_str_action(action)
        return action

    @_launch
    async def custom_send(
        self,
        content: Union[str, CQMsgDict, List[CQMsgDict]],
        isPrivate: bool,
        userId: int = None,
        groupId: int = None,
        cq_str: bool = False,
        waitResp: bool = False,
    ) -> Union[RespEvent, None]:
        """
        自定义发送消息。
        cq_str 若开启，文本中若包含 cq 字符串，将会被解释
        """
        if isPrivate and userId is None:
            raise BotActionError("为私聊时，构建 action 必须提供 userId 参数")
        if not isPrivate and groupId is None:
            raise BotActionError("为群聊时，构建 action 必须提供 groupId 参数")
        action = msg_action(content, isPrivate, userId, groupId, waitResp, self.event)
        if cq_str:
            action = to_cq_str_action(action)
        return action

    @_launch
    async def send_forward(
        self,
        msgNodes: List[MsgNodeDict],
        cq_str: bool = False,
        waitResp: bool = False,
    ) -> Union[RespEvent, None]:
        """
        在当前 session 上下文下发送转发消息。
        cq_str 若开启，文本中若包含 cq 字符串，将会被解释
        """
        if self.event == None:
            raise BotSessionError(
                "空 session 需要使用 custom_send_forward 方法替代 send_forward"
            )
        action = forward_msg_action(
            msgNodes,
            self.event.is_private(),
            self.event.sender.id,
            self.event.group_id,
            waitResp,
            self.event,
        )
        if cq_str:
            action = to_cq_str_action(action)
        return action

    @_launch
    async def custom_send_forward(
        self,
        msgNodes: List[MsgNodeDict],
        isPrivate: bool,
        userId: int = None,
        groupId: int = None,
        cq_str: bool = False,
        waitResp: bool = False,
    ) -> Union[RespEvent, None]:
        """
        自定义发送转发消息。
        cq_str 若开启，文本中若包含 cq 字符串，将会被解释
        """
        action = forward_msg_action(
            msgNodes, isPrivate, userId, groupId, waitResp, self.event
        )
        if cq_str:
            action = to_cq_str_action(action)
        return action

    @_launch
    async def recall(
        self, msgId: int, waitResp: bool = False
    ) -> Union[RespEvent, None]:
        """
        撤回消息
        """
        return msg_del_action(msgId, waitResp, self.event)

    @_launch
    async def get_msg(self, msgId: int) -> RespEvent:
        """
        获取消息信息
        """
        return get_msg_action(msgId, True, self.event)

    @_launch
    async def get_forward_msg(
        self,
        forwardId: str,
    ) -> RespEvent:
        """
        获取转发消息信息
        """
        return get_forward_msg_action(forwardId, True, self.event)

    @_launch
    async def mark_read(
        self, msgId: int, waitResp: bool = False
    ) -> Union[RespEvent, None]:
        """
        标记为已读
        """
        return mark_msg_read_action(msgId, waitResp, self.event)

    @_launch
    async def group_kick(
        self,
        groupId: int,
        userId: int,
        laterReject: bool = False,
        waitResp: bool = False,
    ) -> Union[RespEvent, None]:
        """
        群组踢人
        """
        return group_kick_action(groupId, userId, laterReject, waitResp, self.event)

    @_launch
    async def group_ban(
        self,
        groupId: int,
        userId: int,
        duration: int,
        waitResp: bool = False,
    ) -> Union[RespEvent, None]:
        """
        群组禁言。
        duration 为 0 取消禁言
        """
        return group_ban_action(groupId, userId, duration, waitResp, self.event)

    @_launch
    async def group_ban_anonymous(
        self,
        groupId: int,
        anonymFlag: str,
        duration: int,
        waitResp: bool = False,
    ) -> Union[RespEvent, None]:
        """
        群组匿名禁言。
        无法取消禁言
        """
        return group_anonym_ban_action(
            groupId, anonymFlag, duration, waitResp, self.event
        )

    @_launch
    async def group_ban_all(
        self,
        groupId: int,
        enable: bool,
        waitResp: bool = False,
    ) -> Union[RespEvent, None]:
        """
        群组全员禁言
        """
        return group_whole_ban_action(groupId, enable, waitResp, self.event)

    @_launch
    async def group_leave(
        self,
        groupId: int,
        isDismiss: bool,
        waitResp: bool = False,
    ) -> Union[RespEvent, None]:
        """
        退出群组
        """
        return group_leave_action(groupId, isDismiss, waitResp, self.event)

    @_launch
    async def group_sign(
        self,
        groupId: int,
        waitResp: bool = False,
    ) -> Union[RespEvent, None]:
        """
        群组打卡
        """
        return group_sign_action(groupId, waitResp, self.event)

    @_launch
    async def get_group(
        self,
        groupId: int,
        noCache: bool,
    ) -> RespEvent:
        """
        获取群信息
        """
        return get_group_info_action(groupId, noCache, True, self.event)

    @_launch
    async def get_groups(self) -> RespEvent:
        """
        获取 bot 加入的群列表
        """
        return get_group_list_action(True, self.event)

    @_launch
    async def get_group_member(
        self,
        groupId: int,
        userId: int,
        noCache: bool,
    ) -> RespEvent:
        """
        获取群内单独一个群成员信息
        """
        return get_group_member_info_action(groupId, userId, noCache, True, self.event)

    @_launch
    async def get_group_members(
        self,
        groupId: int,
        noCache: bool,
    ) -> RespEvent:
        """
        获取群成员列表
        """
        return get_group_member_list_action(groupId, noCache, True, self.event)

    @_launch
    async def get_group_honor(
        self,
        groupId: int,
        type: Literal[
            "talkative", "performer", "legend", "strong_newbie", "emotion", "all"
        ],
    ) -> RespEvent:
        """
        获取群荣誉信息
        """
        return get_group_honor_action(groupId, type, True, self.event)

    @_launch
    async def get_group_file_sys(
        self,
        groupId: int,
    ) -> RespEvent:
        """
        获取群文件系统信息
        """
        return get_group_filesys_info_action(groupId, True, self.event)

    @_launch
    async def get_group_root_files(
        self,
        groupId: int,
    ) -> RespEvent:
        """
        获取群根目录文件列表
        """
        return get_group_root_files_action(groupId, True, self.event)

    @_launch
    async def get_group_files_in_folder(self, groupId: int, folderId: str) -> RespEvent:
        """
        获取群子目录文件列表
        """
        return get_group_files_byfolder_action(groupId, folderId, True, self.event)

    @_launch
    async def get_group_file_url(
        self, groupId: int, fileId: str, fileTypeId: int
    ) -> RespEvent:
        """
        获取群文件资源链接。文件相关信息通过 `get_group_root_files` 或
        `get_group_files` 的响应获得
        """
        return get_group_file_url_action(groupId, fileId, fileTypeId, True, self.event)

    @_launch
    async def get_group_sys_msg(self) -> RespEvent:
        """
        获取群系统消息
        """
        return get_group_sys_msg_action(True, self.event)

    @_launch
    async def get_group_notices(
        self,
        groupId: int,
    ) -> RespEvent:
        """
        获取群公告。
        群公告图片有 id，但暂时没有下载的方法
        """
        return get_group_notice_action(groupId, True, self.event)

    @_launch
    async def get_group_records(self, msgSeq: int, groupId: int) -> RespEvent:
        """
        获取群消息历史记录
        """
        return get_group_msg_history_action(msgSeq, groupId, True, self.event)

    @_launch
    async def get_group_essences(self, groupId: int) -> RespEvent:
        """
        获取精华消息列表
        """
        return get_group_essence_list_action(groupId, True, self.event)

    @_launch
    async def set_group_admin(
        self, groupId: int, userId: int, enable: bool, waitResp: bool = False
    ) -> Union[RespEvent, None]:
        """
        设置群管理员
        """
        return set_group_admin_action(groupId, userId, enable, waitResp, self.event)

    @_launch
    async def set_group_card(
        self, groupId: int, userId: int, card: str, waitResp: bool = False
    ) -> Union[RespEvent, None]:
        """
        设置群名片
        """
        return set_group_card_action(groupId, userId, card, waitResp, self.event)

    @_launch
    async def set_group_name(
        self, groupId: int, name: str, waitResp: bool = False
    ) -> Union[RespEvent, None]:
        """
        设置群名
        """
        return set_group_name_action(groupId, name, waitResp, self.event)

    @_launch
    async def set_group_title(
        self,
        groupId: int,
        userId: int,
        title: str,
        duration: int = -1,
        waitResp: bool = False,
    ) -> Union[RespEvent, None]:
        """
        设置群头衔
        """
        return set_group_title_action(
            groupId, userId, title, duration, waitResp, self.event
        )

    @_launch
    async def process_group_add(
        self,
        addFlag: str,
        addType: Literal["add", "invite"],
        approve: bool,
        rejectReason: str = None,
        waitResp: bool = False,
    ) -> Union[RespEvent, None]:
        """
        处理加群请求
        """
        return set_group_add_action(
            addFlag, addType, approve, rejectReason, waitResp, self.event
        )

    @_launch
    async def set_group_icon(
        self, groupId: int, file: str, cache: Literal[0, 1] = 0, waitResp: bool = False
    ) -> Union[RespEvent, None]:
        """
        设置群头像。file 参数接受本地或网络 url 和 base64 编码。
        如本地路径为：`file:///C:/Users/Richard/Pictures/1.png`。
        特别注意：目前此 API 在登录一段时间后会因 cookie 失效而失效
        """
        return set_group_portrait_action(groupId, file, cache, waitResp, self.event)

    @_launch
    async def set_group_notice(
        self, groupId: int, content: str, imageUrl: str = None, waitResp: bool = False
    ) -> Union[RespEvent, None]:
        """
        发送群公告。
        注意 `imageUrl` 只能为本地 url，示例：`file:///C:/users/15742/desktop/123.jpg`
        """
        return set_group_notice_action(groupId, content, imageUrl, waitResp, self.event)

    @_launch
    async def set_group_essence(
        self, msgId: int, type: Literal["add", "del"], waitResp: bool = False
    ) -> Union[RespEvent, None]:
        """
        设置群精华消息
        """
        return set_group_essence_action(msgId, type, waitResp, self.event)

    @_launch
    async def create_group_folder(
        self, groupId: int, folderName: str, waitResp: bool = False
    ) -> Union[RespEvent, None]:
        """
        创建群文件夹。注意：只能在根目录创建文件夹
        """
        return create_group_folder_action(groupId, folderName, waitResp, self.event)

    @_launch
    async def delete_group_folder(
        self, groupId: int, folderId: str, waitResp: bool = False
    ) -> Union[RespEvent, None]:
        """
        删除群文件夹
        """
        return delete_group_folder_action(groupId, folderId, waitResp, self.event)

    @_launch
    async def delete_group_file(
        self, groupId: int, fileId: str, fileTypeId: int, waitResp: bool = False
    ) -> Union[RespEvent, None]:
        """
        删除群文件。文件相关信息通过 `get_group_root_files` 或
        `get_group_files` 的响应获得
        """
        return delete_group_file_action(
            groupId, fileId, fileTypeId, waitResp, self.event
        )

    @_launch
    async def get_friends(self) -> RespEvent:
        """
        获取好友列表
        """
        return get_friend_list_action(True, self.event)

    @_launch
    async def get_undirect_friends(self) -> RespEvent:
        """
        获取单向好友列表
        """
        return get_undirect_friend_action(True, self.event)

    @_launch
    async def get_user(
        self,
        userId: int,
        noCache: bool,
    ) -> RespEvent:
        """
        获取用户信息。可以对陌生人或好友使用
        """
        return get_stranger_info_action(userId, noCache, True, self.event)

    @_launch
    async def process_friend_add(
        self, addFlag: str, approve: bool, remark: str, waitResp: bool = False
    ) -> Union[RespEvent, None]:
        """
        处理加好友。注意 remark 目前暂未实现
        """
        return set_friend_add_action(addFlag, approve, remark, waitResp, self.event)

    @_launch
    async def delete_friend(
        self, userId: int, waitResp: bool = False
    ) -> Union[RespEvent, None]:
        """
        删除好友
        """
        return delete_friend_action(userId, waitResp, self.event)

    @_launch
    async def delete_undirect_friend(
        self, userId: int, waitResp: bool = False
    ) -> Union[RespEvent, None]:
        """
        删除单向好友
        """
        return delete_undirect_friend_action(userId, waitResp, self.event)

    @_launch
    async def get_login_info(self) -> RespEvent:
        """
        获得登录号信息
        """
        return get_login_info_action(True, self.event)

    @_launch
    async def set_login_profile(
        self,
        nickname: str,
        company: str,
        email: str,
        college: str,
        personalNote: str,
        waitResp: bool = False,
    ) -> Union[RespEvent, None]:
        """
        设置登录号资料
        """
        return set_login_profile_action(
            nickname, company, email, college, personalNote, waitResp, self.event
        )

    @_launch
    async def check_send_image(self) -> RespEvent:
        """
        检查是否可以发送图片
        """
        return check_send_image_action(True, self.event)

    @_launch
    async def check_send_audio(self) -> RespEvent:
        """
        检查是否可以发送语音
        """
        return check_send_record_action(True, self.event)

    @_launch
    async def get_cq_status(self) -> RespEvent:
        """
        获取 go-cqhttp 状态
        """
        return get_cq_status_action(True, self.event)

    @_launch
    async def get_cq_version(self) -> RespEvent:
        """
        获取 go-cqhttp 版本信息
        """
        return get_cq_version_action(True, self.event)

    @_launch
    async def quick_handle(
        self,
        contextEvent: BotEvent,
        operation: dict,
        waitResp: bool = False,
    ) -> Union[RespEvent, None]:
        """
        事件快速操作（该方法下一版本实现，本版本无法使用）
        """
        raise ReferenceError("该方法下一版本实现，本版本无法使用")
        # return quick_handle_action(
        #     contextEvent,
        #     operation,
        #     waitResp,
        #     self.event
        # )

    @_launch
    async def get_image(self, fileName: str) -> RespEvent:
        """
        获取图片信息
        """
        return get_image_action(fileName, True, self.event)

    @_launch
    async def download_file(
        self,
        fileUrl: str,
        useThreadNum: int,
        headers: Union[List, str],
        waitResp: bool = True,
    ) -> Union[RespEvent, None]:
        """
        下载文件到缓存目录 action 构造方法。`headers` 的两种格式：
        ```
        "User-Agent=YOUR_UA[\\r\\n]Referer=https://www.baidu.com"
        ```
        或
        ```python
        [
            "User-Agent=YOUR_UA",
            "Referer=https://www.baidu.com"
        ]
        ```
        """
        return download_file_action(
            fileUrl, useThreadNum, headers, waitResp, self.event
        )

    @_launch
    async def ocr(
        self,
        image: str,
    ) -> RespEvent:
        """
        图片 OCR。image 为图片 ID
        """
        return ocr_action(image, True, self.event)

    @_launch
    async def upload_file(
        self,
        isPrivate: bool,
        file: str,
        sendFileName: str,
        userId: int = None,
        groupId: int = None,
        groupFolderId: str = None,
        waitResp: bool = False,
    ) -> Union[RespEvent, None]:
        """
        发送文件 action 构造方法。只支持发送本地文件。
        若为群聊文件发送，不提供 folder id，则默认上传到群文件根目录。

        示例路径：`C:/users/15742/desktop/QQ图片20230108225606.jpg`。

        （若需要发送网络文件，先使用 `download_file` 方法下载网络文件。
        响应后文件会放于 go-cqhttp 缓存文件夹中，可直接在消息段中引用）
        """
        return upload_file_action(
            isPrivate,
            file,
            sendFileName,
            userId,
            groupId,
            groupFolderId,
            waitResp,
            self.event,
        )

    @_launch
    async def get_at_all_remain(self, groupId: int) -> RespEvent:
        """
        获取群 @全体成员 剩余次数
        """
        return get_atall_remain_action(groupId, True, self.event)

    @_launch
    async def get_online_clients(
        self,
        noCache: bool,
    ) -> RespEvent:
        """
        获取当前账号在线客户端列表
        """
        return get_online_clients_action(noCache, True, self.event)

    @_launch
    async def get_model_show(
        self,
        model: str,
    ) -> RespEvent:
        """
        获取在线机型
        """
        return get_model_show_action(model, True, self.event)

    @_launch
    async def set_model_show(self, model: str, modelShow: str) -> RespEvent:
        """
        设置在线机型
        """
        return set_model_show_action(model, modelShow, True, self.event)


class BotSessionManager:
    STORAGE: Dict[object, Set[BotSession]] = {}
    HUP_STORAGE: Dict[object, Set[BotSession]] = {}
    # 各个 handler 对饮的操作锁
    WORK_LOCKS: Dict[object, aio.Lock] = {}
    # 用来标记 cls.get 等待一个挂起的 session 时的死锁
    DEADLOCK_FLAGS: Dict[object, aio.Event] = {}
    # 对应每个 handler 的 try_attach 过程的操作锁
    ATTACH_LOCKS: Dict[object, aio.Lock] = {}

    @classmethod
    def register(cls, handler: object) -> None:
        """
        以 handler 为键，注册 handler 对应的 session 空间、操作锁和挂起 session 空间
        """
        cls.STORAGE[handler] = set()
        cls.WORK_LOCKS[handler] = aio.Lock()
        cls.HUP_STORAGE[handler] = set()
        cls.DEADLOCK_FLAGS[handler] = aio.Event()
        cls.ATTACH_LOCKS[handler] = aio.Lock()

    @classmethod
    def inject(cls, session: BotSession, handler: object) -> None:
        """
        handler 内绑定 handler 引用到 session
        """
        session._handler = handler

    @classmethod
    def __attach(cls, event: BotEvent, handler: object) -> bool:
        """
        session 附着操作，临界区操作。只能在 cls.try_attach 中进行
        """
        session = None
        for s in cls.HUP_STORAGE[handler]:
            # session 的挂起方法，保证 session 一定未过期，因此不进行过期检查
            if handler._rule.compare(s.event, event):
                session = s
                break
        # 如果获得一个挂起的 session，它一定是可附着的，附着后需要唤醒
        if session:
            session.events.append(event)
            cls._rouse(session)
            return True
        return False

    @classmethod
    async def try_attach(cls, event: BotEvent, handler: object) -> bool:
        """
        检查是否有挂起的 session 可供 event 附着。
        如果有则附着并唤醒，并返回 True。否则返回 False。
        """
        if handler._rule is None:
            return False

        async with cls.ATTACH_LOCKS[handler]:
            t1 = aio.create_task(cls.DEADLOCK_FLAGS[handler].wait(), name="flag")
            t2 = aio.create_task(cls.WORK_LOCKS[handler].acquire(), name="lock")
            done, _ = await aio.wait([t1, t2], return_when=aio.FIRST_COMPLETED)
            # 等待完成后，一定要记得取消另一个任务！否则可能异常加锁
            if done.pop().get_name() == "flag":
                res = cls.__attach(event, handler)
                cls.DEADLOCK_FLAGS[handler].clear()
                t2.cancel()
                return res
            else:
                res = cls.__attach(event, handler)
                cls.WORK_LOCKS[handler].release()
                t1.cancel()
                return res

    @classmethod
    def _hup(cls, session: BotSession) -> None:
        """
        挂起 session
        """
        if session._space_tag is None:
            raise BotSessionError(
                "一次性 session 或空 session 不支持挂起，因为缺乏 session_rule 作为唤醒标志"
            )
        elif session._expired:
            raise BotSessionError("过期的 session 不能被挂起")
        session.hup_times.append(time.time())
        cls.STORAGE[session._space_tag].remove(session)
        cls.HUP_STORAGE[session._space_tag].add(session)
        session._awake_signal.clear()

    @classmethod
    def _rouse(cls, session: BotSession) -> None:
        """
        唤醒 session
        """
        cls.HUP_STORAGE[session._space_tag].remove(session)
        cls.STORAGE[session._space_tag].add(session)
        session._awake_signal.set()

    @classmethod
    def _expire(cls, session: BotSession) -> None:
        """
        标记该 session 为过期状态，并进行销毁操作（如果存在于某个 session_space，则从中移除）
        """
        if session._expired:
            return
        session.events.clear()
        session.hup_times.clear()
        session.store_clear()
        session._expired = True
        if session._space_tag:
            cls.STORAGE[session._space_tag].remove(session)

    @classmethod
    def recycle(cls, session: BotSession, alive: bool = False) -> None:
        """
        session 所在方法运行结束后，回收 session。
        默认将 session 销毁。若 alive 为 True，则保留
        """
        session._free_signal.set()
        if not alive:
            cls._expire(session)

    @classmethod
    async def get(
        cls, event: BotEvent, responder: IActionResponder, handler: object
    ) -> Union[BotSession, None]:
        """
        handler 内获取 session 方法。自动根据 handler._rule 判断是否需要映射到 session_space 进行存储。
        然后根据具体情况，获取已有 session 或新建 session。当尝试获取非空闲 session 时，如果 handler 指定不等待则返回 None
        """
        if handler._rule:
            # session_space, session._free_signal 竞争，需要加锁
            async with cls.WORK_LOCKS[handler]:
                session = await cls._get_on_rule(event, responder, handler)
                # 必须在锁的保护下修改 session._free_signal
                if session:
                    session._free_signal.clear()
        else:
            session = cls._make(event, responder, handler)
            session._free_signal.clear()

        return session

    @classmethod
    def _make(
        cls, event: BotEvent, responder: IActionResponder, handler: object = None
    ) -> BotSession:
        """
        内部使用的创建 session 方法。如果 handler 为空，即缺乏 space_tag，则为一次性 session。
        或 handler._rule 为空，则也为一次性 session
        """
        if handler:
            if handler._rule:
                session = BotSession(responder, handler)
                session.events.append(event)
                cls.STORAGE[handler].add(session)
                return session
        session = BotSession(responder)
        session.events.append(event)
        return session

    @classmethod
    def make_empty(cls, responder: IActionResponder) -> BotSession:
        """
        创建空 session。即不含 event 和 space_tag 标记的 session
        """
        return BotSession(responder)

    @classmethod
    def make_temp(cls, event: BotEvent, responder: IActionResponder) -> BotSession:
        """
        创建一次性 session。确定无需 session 管理机制时可以使用。
        否则一定使用 cls.get 方法
        """
        return cls._make(event, responder)

    @classmethod
    async def _get_on_rule(
        cls, event: BotEvent, responder: IActionResponder, handler: object
    ) -> Union[BotSession, None]:
        """
        根据 handler 具体情况，从对应 session_space 中获取 session 或新建 session。
        或从 hup_session_space 中唤醒 session，或返回 None
        """
        session = None
        check_rule, session_space, hup_session_space, conflict_wait = (
            handler._rule,
            cls.STORAGE[handler],
            cls.HUP_STORAGE[handler],
            handler._wait_flag,
        )

        # for 循环都需要即时 break，保证遍历 session_space 时没有协程切换。因为切换后 session_space 可能发生变动
        for s in session_space:
            if check_rule.compare(s.event, event) and not s._expired:
                session = s
                break
        # 如果会话不存在，生成一个新 session 变量
        if session is None:
            return cls._make(event, responder, handler)
        # 如果会话存在，且未过期，且空闲，则附着到这个 session 上
        if session._free_signal.is_set():
            session.events.append(event)
            return session
        # 如果会话存在，且未过期，但是不空闲，选择不等待
        if not conflict_wait:
            return None
        # 如果会话存在，且未过期，但是不空闲，选择等待，此时就不得不陷入等待（即将发生协程切换）
        await aio.wait(
            [session._free_signal.wait(), session._hup_signal.wait()],
            return_when=aio.FIRST_COMPLETED,
        )
        if session._hup_signal.is_set():
            cls.DEADLOCK_FLAGS[handler].set()
            await session._free_signal.wait()
        """
        重新切换回本协程后，session 有可能变为过期，但此时一定是空闲的。
        同时一定是非挂起状态。因为上面解决了可能存在的挂起死锁问题。
        即使该 session 因过期被所在的 session_space 清除也无妨，因为此处有引用，
        该 session 并不会消失。且此处不操作 session_space，无需担心 session_space 变动
        """
        # 如果过期，生成一个新的 session 变量
        if session._expired:
            return cls._make(event, responder, handler)
        # 如果未过期，则附着到这个 session 上
        else:
            session.events.append(event)
            return session


_session_ctx = ContextVar("session_ctx")


class SessionLocal:
    """
    session 自动上下文
    """

    __slots__ = tuple(
        list(filter(lambda x: not (len(x) >= 2 and x[:2] == "__"), dir(BotSession)))
        + ["__storage__"]
    )

    def __init__(self) -> None:
        object.__setattr__(self, "__storage__", _session_ctx)
        self.__storage__: ContextVar[BotSession]

    def __setattr__(self, __name: str, __value: Any) -> None:
        setattr(self.__storage__.get(), __name, __value)

    def __getattr__(self, __name: str) -> Any:
        return getattr(self.__storage__.get(), __name)

    def _add_ctx(self, ctx: BotSession) -> Token:
        return self.__storage__.set(ctx)

    def _del_ctx(self, token: Token) -> None:
        self.__storage__.reset(token)


class AttrSessionRule(SessionRule):
    def __init__(self, *attrs) -> None:
        super().__init__()
        self.attrs = attrs

    def _get_val(self, e: BotEvent, attrs: Tuple[str, ...]) -> Any:
        val = e
        try:
            for attr in attrs:
                val = getattr(val, attr)
        except AttributeError:
            raise BotSessionError(f"session 规则指定的属性 {attr} 不存在")
        return val

    def compare(self, e1: BotEvent, e2: BotEvent) -> bool:
        return self._get_val(e1, self.attrs) == self._get_val(e2, self.attrs)


SESSION_LOCAL = SessionLocal()
SESSION_LOCAL: BotSession


async def send(
    content: Union[str, CQMsgDict, List[CQMsgDict]],
    cq_str: bool = False,
    wait: bool = False,
) -> Union[RespEvent, None]:
    """
    发送一条消息
    """
    return await SESSION_LOCAL.send(content, cq_str, wait)


async def send_hup(
    content: Union[str, CQMsgDict, List[CQMsgDict]],
    cq_str: bool = False,
    overtime: int = None,
) -> None:
    """
    回复一条消息然后挂起
    """
    await SESSION_LOCAL.send(content, cq_str)
    await SESSION_LOCAL.suspend(overtime)


async def send_reply(
    content: Union[str, CQMsgDict, List[CQMsgDict]],
    cq_str: bool = False,
    wait: bool = False,
) -> Union[RespEvent, None]:
    """
    发送一条消息（以回复消息的形式，回复触发动作的那条消息）
    """
    content_arr = [reply_msg(SESSION_LOCAL.event.id)]
    if isinstance(content, str):
        content_arr.append(text_msg(content))
    elif isinstance(content, dict):
        content_arr.append(content)
    else:
        content_arr.extend(content)
    return await SESSION_LOCAL.send(content_arr, cq_str, wait)


async def finish(
    content: Union[str, CQMsgDict, List[CQMsgDict]], cq_str: bool = False
) -> None:
    """
    回复一条消息，然后直接结束当前事件处理方法
    """
    await SESSION_LOCAL.send(content, cq_str)
    SESSION_LOCAL.destory()
    raise DirectRetSignal("事件处理方法被安全地递归 return，请无视这个异常")


def event():
    """
    获取当前 session 上下文下的 event
    """
    return SESSION_LOCAL.event


def msg_text():
    """
    获取当前 session 上下文下的 event.text，如果不存在则返回 None
    """
    if hasattr(SESSION_LOCAL.event, "text"):
        return SESSION_LOCAL.event.text
    return None



import traceback
from abc import abstractmethod

from ..types.exceptions import *
from ..types.typing import *


class TipGenerator:
    """
    提示信息生成器
    """

    @abstractmethod
    def gen(
        self,
        src: str,
        src_desc: Union[str, None],
        src_expect: Union[str, None],
        idx: int,
        exc_type: Exception,
        exc_tb: traceback,
    ) -> str:
        pass


class ArgFormatter:
    """
    参数格式化器。将参数格式化为对应类型的对象。格式化后可添加校验，也可自定义失败提示信息。
    还可设置默认值。不提供默认值参数时为 Null，代表不使用默认值
    """

    def __init__(
        self,
        convert: Callable[[str], Any] = None,
        verify: Callable[[Any], bool] = None,
        src_desc: str = None,
        src_expect: str = None,
        default: Any = Null,
        default_replace_flag: str = None,
        convert_tip_gen: TipGenerator = None,
        verify_tip_gen: TipGenerator = None,
        arglack_tip_gen: TipGenerator = None,
    ) -> None:
        self.convert = convert
        self.verify = verify
        self.src_desc = src_desc
        self.src_expect = src_expect
        self.default = default
        self.default_replace_flag = default_replace_flag
        if self.default is Null and self.default_replace_flag is not None:
            raise ArgFormatInitError(
                "初始化参数格式化器时，使用“默认值替换标记”必须同时设置默认值"
            )

        self.out_convert_tip_gen = convert_tip_gen
        self.out_verify_tip_gen = verify_tip_gen
        self.out_arglack_tip_gen = arglack_tip_gen

    def _get_val(self, args: ParseArgs, idx: int) -> Any:
        if self.default is Null:
            if args.vals is None or len(args.vals) < idx + 1:
                raise ArgLackError
            else:
                return args.vals[idx]
        if args.vals is None:
            args.vals = [self.default]
        elif len(args.vals) < idx + 1:
            args.vals.append(self.default)
        return args.vals[idx]

    def format(self, args: ParseArgs, idx: int) -> None:
        """
        格式化参数为对应类型的变量
        """
        try:
            src = self._get_val(args, idx)
            if (
                self.default_replace_flag is not None
                and src == self.default_replace_flag
            ):
                src = self.default
            res = self.convert(src) if self.convert is not None else src

            if self.verify is None:
                pass
            elif self.verify(res):
                pass
            else:
                raise ArgVerifyFailed
            args.vals[idx] = res
        except ArgVerifyFailed as e:
            if self.out_verify_tip_gen:
                tip = self.out_verify_tip_gen.gen(
                    src, self.src_desc, self.src_expect, idx, e, traceback
                )
            else:
                tip = self._verify_tip_gen(src, idx)
            raise ArgFormatFailed(tip)
        except ArgLackError as e:
            if self.out_arglack_tip_gen:
                tip = self.out_arglack_tip_gen.gen(
                    None, self.src_desc, self.src_expect, idx, e, traceback
                )
            else:
                tip = self._arglack_tip_gen(idx)
            raise ArgFormatFailed(tip)
        except Exception as e:
            if self.out_convert_tip_gen:
                tip = self.out_convert_tip_gen.gen(
                    src, self.src_desc, self.src_expect, idx, e, traceback
                )
            else:
                tip = self._convert_tip_gen(src, idx, e)
            raise ArgFormatFailed(tip)

    def _convert_tip_gen(self, src: str, idx: int, exc_type: Exception) -> str:
        e_class = exc_type.__class__.__name__
        src = src.__repr__() if isinstance(src, str) else src
        tip = f"第 {idx+1} 个参数"
        tip += (
            f"（{self.src_desc}）无法处理，给定的值为：{src}。"
            if self.src_desc
            else f"给定的值 {src} 无法处理。"
        )
        tip += f"参数要求：{self.src_expect}。" if self.src_expect else ""
        tip += f"\n详细错误描述：[{e_class}] {exc_type}"
        return tip

    def _verify_tip_gen(self, src: str, idx: int) -> str:
        src = src.__repr__() if isinstance(src, str) else src
        tip = f"第 {idx+1} 个参数"
        tip += (
            f"（{self.src_desc}）不符合要求，给定的值为：{src}。"
            if self.src_desc
            else f"给定的值 {src} 不符合要求。"
        )
        tip += f"参数要求：{self.src_expect}。" if self.src_expect else ""
        return tip

    def _arglack_tip_gen(self, idx: int) -> str:
        tip = f"第 {idx+1} 个参数"
        tip += f"（{self.src_desc}）缺失。" if self.src_desc else f"缺失。"
        tip += f"参数要求：{self.src_expect}。" if self.src_expect else ""
        return tip



from abc import ABC, abstractmethod
from asyncio import Future

from ..models.action import BotAction
from ..models.event import BotEvent, MetaEvent, RespEvent


class IRespDispatcher(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    async def dispatch(self, resp: RespEvent) -> None:
        pass


class IActionSender(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    async def send(self, action: BotAction) -> None:
        pass


class IActionResponder(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    async def take_action(self, action: BotAction) -> None:
        pass

    @abstractmethod
    async def take_action_wait(self, action: BotAction) -> Future[RespEvent]:
        pass


class IEventDispatcher(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    async def dispatch(self, event: BotEvent) -> None:
        pass



import asyncio as aio
import re
from contextlib import asynccontextmanager
from itertools import chain, zip_longest

from ..types.exceptions import *
from ..types.typing import *


class Singleton:
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "__instance__"):
            cls.__instance__ = super(Singleton, cls).__new__(cls)
        return cls.__instance__


def to_cq_arr(s: str) -> dict:
    """
    从 cq 消息字符串转换为 cq 消息段
    """

    def replace_func(m) -> str:
        s, e = m.regs[0]
        cq_texts.append(m.string[s:e])
        return "\u0000"

    cq_regex = re.compile(r"\[CQ:.*?\]")
    cq_texts = []
    no_cq_str = cq_regex.sub(replace_func, s)
    pure_texts = map(
        lambda x: f"[CQ:text,text={x}]" if x != "" else x,
        no_cq_str.split("\u0000"),
    )
    content: str = "".join(
        chain.from_iterable(zip_longest(pure_texts, cq_texts, fillvalue=""))
    )

    cq_entity = content.split("]")[:-1]
    content: list = []
    for e in cq_entity:
        _ = e.split(",")
        cq_type = _[0][4:]
        data = {}
        for param_pair in _[1:]:
            name, val = param_pair.split("=")
            val = (
                val.replace("&amp;", "&")
                .replace("&#91;", "[")
                .replace("&#93;", "]")
                .replace("&#44;", ",")
            )
            if val.isdigit():
                val = int(val)
            try:
                val = float(val)
            except Exception:
                pass
            data[name] = val
        content.append({"type": cq_type, "data": data})
    return content


class AsyncTwinEvent(aio.Event):
    """
    孪生 Event，会和绑定的一方时刻保持状态相反。
    """

    def __init__(self) -> None:
        super().__init__()
        self._twin: AsyncTwinEvent = None

    def bind(self, twin: "AsyncTwinEvent") -> None:
        self._twin = twin
        if self.is_set():
            super(AsyncTwinEvent, self._twin).clear()
        else:
            super(AsyncTwinEvent, self._twin).set()

    def set(self) -> None:
        super().set()
        if self._twin:
            super(AsyncTwinEvent, self._twin).clear()

    def clear(self) -> None:
        super().clear()
        if self._twin:
            super(AsyncTwinEvent, self._twin).set()


def get_twin_event() -> Tuple[aio.Event, aio.Event]:
    """
    获得两个时刻保持状态相反的 asyncio.Event。
    获得的第一个为 unset，另一个为 set
    """
    a, b = AsyncTwinEvent(), AsyncTwinEvent()
    a.bind(b)
    b.bind(a)
    return a, b


class RWController:
    """
    异步读写控制器。提供异步安全的读写上下文
    """

    def __init__(self, read_limit: int = None) -> None:
        @asynccontextmanager
        async def safe_read():
            nonlocal read_num, read_semaphore, write_semaphore, read_num_lock
            if read_semaphore:
                await read_semaphore.acquire()
            async with read_num_lock:
                if read_num == 0:
                    await write_semaphore.acquire()
                read_num += 1
            try:
                yield
            finally:
                async with read_num_lock:
                    read_num -= 1
                    if read_num == 0:
                        write_semaphore.release()
                    if read_semaphore:
                        read_semaphore.release()

        @asynccontextmanager
        async def safe_write():
            nonlocal write_semaphore
            await write_semaphore.acquire()
            try:
                yield
            finally:
                write_semaphore.release()

        write_semaphore = aio.Semaphore(1)
        if read_limit:
            read_semaphore = aio.Semaphore(read_limit)
        else:
            read_semaphore = None
        read_num = 0
        read_num_lock = aio.Lock()
        self.safe_read = safe_read
        self.safe_write = safe_write



from abc import ABC, abstractmethod, abstractproperty

from .exceptions import *
from .typing import *


class BotEvent(ABC):
    """
    Bot 事件类
    """

    def __init__(self, rawEvent: dict) -> None:
        self.raw = rawEvent
        self._args_map: Dict[Any, Dict[str, ParseArgs]] = None

    @abstractproperty
    def time(self) -> int:
        pass

    @abstractproperty
    def type(self) -> str:
        pass

    def is_msg(self) -> bool:
        return self.type == "message"

    def is_req(self) -> bool:
        return self.type == "request"

    def is_notice(self) -> bool:
        return self.type == "notice"

    def is_meta(self) -> bool:
        return self.type == "meta"

    def is_resp(self) -> bool:
        return self.type == "response"

    def _get_args(self, parser_id: Any) -> Union[Dict[str, ParseArgs], Literal[-1]]:
        if self._args_map is None:
            return -1
        return self._args_map.get(parser_id, -1)

    def _store_args(self, parser_id: Any, args_group: Dict[str, ParseArgs]) -> None:
        if self._args_map is None:
            self._args_map = {}
        self._args_map[parser_id] = args_group


class SessionRule(ABC):
    """
    用作 sesion 的区分依据
    """

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def compare(self, e1: BotEvent, e2: BotEvent) -> bool:
        pass


class BotLife(Enum):
    """
    bot 生命周期枚举
    """

    LOADED = 1
    CONNECTED = 2
    BEFORE_CLOSE = 3
    BEFORE_STOP = 4
    EVENT_RECEIVED = 5
    ACTION_PRESEND = 6


# 插件共享对象构造参数
ShareObjArgs = NamedTuple("ShareObjArgs", property=str, namespace=str, id=str)
# 插件共享对象回调的构造参数
ShareCbArgs = NamedTuple("ShareCbArgs", namespace=str, id=str, cb=Callable)
# 插件信号方法构造参数
SignalHandlerArgs = NamedTuple(
    "SignalHandlerArgs", func=AsyncFunc[None], namespace=str, signal=str
)
# 钩子方法（生命周期回调）构造参数
HookRunnerArgs = NamedTuple("HookRunnerArgs", func=AsyncFunc[None], type=BotLife)



from .action import *
from .base import RWController, get_twin_event, to_cq_arr
from .bot import BOT_PROXY as bot
from .event import MetaEvent, MsgEvent, NoticeEvent, RequestEvent, RespEvent
from .ipc import PluginBus, PluginStore
from .plugin import Plugin
from .session import SESSION_LOCAL as session
from .session import (
    AttrSessionRule,
    BotSession,
    event,
    finish,
    msg_text,
    send,
    send_hup,
    send_reply,
)



import logging
import logging.config
import logging.handlers
import os
from logging import CRITICAL, DEBUG, ERROR, INFO, WARN, WARNING, Logger

from ..types.typing import *


def get_logger(
    log_dir: str = None,
    log_level: Literal[
        "DEBUG", "ERROR", "INFO", "WARN", "WARNING", "CRITICAL"
    ] = "DEBUG",
) -> Logger:
    """
    无日志目录时获取只含 console 输出的 logger，否则返回含文件输出的 logger
    """
    if log_dir:
        if not os.path.exists(log_dir):
            os.mkdir(log_dir)
        config = get_config(log_dir, log_level)
    else:
        config = get_fileless_config(log_level)
    logging.config.dictConfig(config)
    return logging.getLogger("bot_logger")


def get_fileless_config(log_level: str):
    """
    返回只含 console 输出的 logger
    """
    config = LOG_CONFIG.copy()
    config["handlers"]["console_handler"]["level"] = LOG_LEVEL_MAP[log_level]
    return config


def get_config(log_dir: str, log_level: str):
    """
    返回含 console 输出和文件输出的 logger
    """
    config = LOG_CONFIG.copy()
    config["handlers"]["console_handler"]["level"] = LOG_LEVEL_MAP[log_level]
    config["handlers"]["file_handler"] = {
        "class": "logging.handlers.RotatingFileHandler",
        "level": logging.DEBUG,
        "formatter": "file_formatter",
        "filename": os.path.join(log_dir, "bot.log"),
        "maxBytes": 1024 * 1024,
        "backupCount": 10,
        "encoding": "UTF-8",
    }
    config["loggers"]["bot_logger"]["handlers"].append("file_handler")
    return config


LOG_COLOR_CONFIG = {
    "DEBUG": "purple",
    "INFO": "white",
    "WARNING": "yellow",
    "ERROR": "red",
    "CRITICAL": "bold_red",
}


LOG_LEVEL_MAP = {
    "DEBUG": DEBUG,
    "INFO": INFO,
    "WARN": WARN,
    "WARNING": WARNING,
    "ERROR": ERROR,
    "CRITICAL": CRITICAL,
}


LOG_CONFIG = {
    "version": 1,
    "disable_existing_loggers": True,
    "incremental": False,
    "formatters": {
        "console_formatter": {
            "()": "colorlog.ColoredFormatter",
            "format": "%(log_color)s[%(asctime)s] [%(levelname)s]: %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
            "log_colors": LOG_COLOR_CONFIG,
        },
        "file_formatter": {
            "class": "logging.Formatter",
            "format": "[%(asctime)s] [%(filename)s %(lineno)d] [%(levelname)s]: %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        },
    },
    "handlers": {
        "console_handler": {
            "class": "logging.StreamHandler",
            "level": logging.INFO,
            "formatter": "console_formatter",
        }
    },
    "loggers": {
        "bot_logger": {
            "handlers": ["console_handler"],
            "level": logging.DEBUG,
            "propagate": False,
        }
    },
}



import json
from abc import ABC
from copy import deepcopy

from ..types.exceptions import *
from ..types.typing import *
from ..utils.base import get_id
from .event import *

__all__ = (
    "BotAction",
    # 消息构造方法
    "text_msg",
    "face_msg",
    "audio_msg",
    "at_msg",
    "share_msg",
    "music_msg",
    "custom_music_msg",
    "image_msg",
    "reply_msg",
    "touch_msg",
    "tts_msg",
    # cq 编码相关
    "cq_escape",
    "cq_anti_escape",
    "to_cq_str_action",
    # 消息发送、处理相关
    "msg_action",
    "forward_msg_action",
    "custom_msg_node",
    "refer_msg_node",
    "msg_del_action",
    "get_msg_action",
    "get_forward_msg_action",
    "mark_msg_read_action",
    # 群相关
    "group_kick_action",
    "group_ban_action",
    "group_anonym_ban_action",
    "group_whole_ban_action",
    "group_leave_action",
    "group_sign_action",
    "get_group_info_action",
    "get_group_list_action",
    "get_group_member_info_action",
    "get_group_member_list_action",
    "get_group_honor_action",
    "get_group_filesys_info_action",
    "get_group_root_files_action",
    "get_group_files_byfolder_action",
    "get_group_file_url_action",
    "get_group_sys_msg_action",
    "get_group_notice_action",
    "get_group_msg_history_action",
    "get_group_essence_list_action",
    "set_group_admin_action",
    "set_group_card_action",
    "set_group_name_action",
    "set_group_title_action",
    "set_group_add_action",
    "set_group_portrait_action",
    "set_group_notice_action",
    "set_group_essence_action",
    "create_group_folder_action",
    "delete_group_folder_action",
    "delete_group_file_action",
    # 私聊相关
    "get_friend_list_action",
    "get_undirect_friend_action",
    "get_stranger_info_action",
    "set_friend_add_action",
    "delete_friend_action",
    "delete_undirect_friend_action",
    # bot 前后端相关
    "get_login_info_action",
    "set_login_profile_action",
    "check_send_image_action",
    "check_send_record_action",
    "get_cq_status_action",
    "get_cq_version_action",
    "quick_handle_action",
    # 其他操作
    "get_image_action",
    "download_file_action",
    "ocr_action",
    "upload_file_action",
    "get_atall_remain_action",
    # 登录设备和机型相关
    "get_online_clients_action",
    "get_model_show_action",
    "set_model_show_action",
)


def cq_escape(text: str) -> str:
    """
    cq 码特殊字符转义
    """
    return (
        text.replace("&", "&amp;")
        .replace("[", "&#91;")
        .replace("]", "&#93;")
        .replace(",", "&#44;")
    )


def cq_anti_escape(text: str) -> str:
    """
    cq 码特殊字符逆转义
    """
    return (
        text.replace("&amp;", "&")
        .replace("&#91;", "[")
        .replace("&#93;", "]")
        .replace("&#44;", ",")
    )


def to_cq_str_action(action: "BotAction") -> "BotAction":
    """
    转化 action 携带的 message 字段转为 cq 字符串格式，并返回新的 action。
    支持的 action 类型有：msg_action 和 forward_action
    """

    def _format_msg_action(action: "BotAction") -> None:
        cq_str_list = []
        for item in action.params["message"]:
            if item["type"] != "text":
                string = (
                    f"[CQ:{item['type']},"
                    + ",".join([f"{k}={item['data'][k]}" for k in item["data"].keys()])
                    + "]"
                )
                cq_str_list.append(string)
            else:
                cq_str_list.append(item["data"]["text"])
        action.params["message"] = "".join(cq_str_list)

    def _format_forward_action(action: "BotAction") -> None:
        for item in action.params["messages"]:
            if "id" in item["data"].keys():
                continue

            cq_str_list = []
            for msg in item["data"]["content"]:
                if msg["type"] != "text":
                    string = (
                        f"[CQ:{msg['type']},"
                        + ",".join(
                            [f"{k}={msg['data'][k]}" for k in msg["data"].keys()]
                        )
                        + "]"
                    )
                    cq_str_list.append(string)
                else:
                    cq_str_list.append(msg["data"]["text"])
            item["data"]["content"] = "".join(cq_str_list)

    _action = deepcopy(action)
    if _action.type == "send_msg":
        _format_msg_action(_action)
    elif _action.type in ("send_private_forward_msg", "send_group_forward_msg"):
        _format_forward_action(_action)
    else:
        raise BotActionError("传入的 action 因类型不匹配，不可被 cq 序列化")

    return _action


def text_msg(
    text: str,
) -> CQMsgDict:
    """
    普通文本消息
    """
    return {"type": "text", "data": {"text": text}}


def face_msg(
    icon_id: int,
) -> CQMsgDict:
    """
    QQ 表情
    """
    return {"type": "face", "data": {"id": f"{icon_id}"}}


def audio_msg(
    url: str,
    timeout: int = None,
    magic: bool = False,
) -> CQMsgDict:
    """
    语音消息
    """
    base = {
        "type": "record",
        "data": {
            "file": url,
        },
    }
    if magic:
        base["data"]["magic"] = 1
    if timeout:
        base["data"]["timeout"] = str(timeout)
    return base


def at_msg(
    qqId: Union[int, Literal["all"]],
    notInName: str = None,
) -> CQMsgDict:
    """
    at 消息。
    at 所有人时，`qqId` 传 "all"
    """
    base = {
        "type": "at",
        "data": {
            "qq": str(qqId),
        },
    }
    if notInName:
        base["data"]["name"] = notInName
    return base


def share_msg(
    url: str,
    title: str,
    content: str = None,
    image: str = None,
) -> CQMsgDict:
    """
    链接分享卡片消息。
    `content` 为描述语
    """
    base = {
        "type": "share",
        "data": {
            "url": url,
            "title": title,
        },
    }
    if content:
        base["data"]["content"] = content
    if image:
        base["data"]["image"] = image
    return base


def music_msg(
    platType: Literal["qq", "163", "xm"],
    songId: str,
) -> CQMsgDict:
    """
    音乐分享卡片消息（专有平台）
    """
    return {"type": "music", "data": {"type": platType, "id": songId}}


def custom_music_msg(
    url: str,
    audio: str,
    title: str,
    content: str = None,
    image: str = None,
) -> CQMsgDict:
    """
    自定义音乐分享卡片。
    `url` 为主页或网站起始页
    """
    base = {
        "type": "music",
        "data": {
            "type": "custom",
            "url": url,
            "audio": audio,
            "title": title,
        },
    }
    if content:
        base["data"]["content"] = content
    if image:
        base["data"]["image"] = image
    return base


def image_msg(
    url: str,
    picType: Literal["flash", "show"] = None,
    subType: Literal[0, 1] = None,
    useCache: Literal[0, 1] = 1,
) -> CQMsgDict:
    """
    图片消息。
    `url`: 图片 url。可以为本地路径，如：`file:///C:/users/15742/desktop/QQ图片20230108225606.jpg`；也可以为网络 url；还可以为 image id。
    `picType`: flash 为闪照，show 为秀图，不填为普通图片。
    `subType`: 只出现在群聊，0 为正常图片，1 为表情包
    """
    base = {
        "type": "image",
        "data": {
            "file": url,
        },
    }
    if picType:
        base["data"]["type"] = picType
    if subType:
        base["data"]["subType"] = subType
    if useCache:
        base["data"]["cache"] = useCache
    return base


def reply_msg(
    messageId: int,
) -> CQMsgDict:
    """
    回复消息
    """
    return {
        "type": "reply",
        "data": {
            "id": messageId,
        },
    }


def touch_msg(
    qqId: int,
) -> CQMsgDict:
    """
    戳一戳消息
    """
    return {
        "type": "poke",
        "data": {
            "qq": qqId,
        },
    }


def touch_msg(
    qqId: int,
) -> CQMsgDict:
    """
    openshamrock 的戳一戳消息
    """
    return {
        "type": "touch",
        "data": {
            "id": qqId,
        },
    }


def tts_msg(
    text: str,
) -> CQMsgDict:
    """
    腾讯自带 tts 语音消息
    """
    return {
        "type": "tts",
        "data": {
            "text": text,
        },
    }


def custom_msg_node(
    content: Union[str, CQMsgDict, List[CQMsgDict]],
    sendName: str,
    sendId: int,
    seq: str = None,
) -> MsgNodeDict:
    """
    自定义消息节点构造方法。转化字符串、消息、消息段为消息节点
    """
    if isinstance(content, str):
        msgs = text_msg(content)
        if not isinstance(msgs, list):
            msgs = [msgs]
    elif isinstance(content, dict):
        msgs = [content]
    elif isinstance(content, list):
        temp = []
        for _ in content:
            if isinstance(_, list):
                temp.extend(_)
            else:
                temp.append(_)
        msgs = temp
    ret = {
        "type": "node",
        "data": {"name": sendName, "uin": str(sendId), "content": msgs},
    }
    if seq:
        ret["data"]["seq"] = seq
    return ret


def refer_msg_node(msgId: int) -> MsgNodeDict:
    """
    引用消息节点构造方法
    """
    return {"type": "node", "data": {"id": str(msgId)}}


class ActionPacker(ABC):
    """
    行为信息构造基类
    """

    def __init__(self) -> None:
        super().__init__()
        self.type: str
        self.params: dict


class BotAction:
    """
    Bot 行为类
    """

    def __init__(
        self,
        package: ActionPacker,
        respWaited: bool = False,
        triggerEvent: BotEvent = None,
    ) -> None:
        # 只有 action 对应的响应需要被等待单独处理时，才会生成 id
        self.resp_id: Union[str, None] = str(get_id()) if respWaited else None
        self.type: str = package.type
        self.params: dict = package.params
        self.trigger: Union[BotEvent, None] = triggerEvent

    def copy(self) -> "BotAction":
        """
        创建当前 action 的深拷贝
        """
        return deepcopy(self)

    def extract(self) -> dict:
        """
        从对象提取标准 cq action dict
        """
        obj = {
            "action": self.type,
            "params": self.params,
        }
        if self.resp_id:
            obj["echo"] = self.resp_id
        return obj

    def flatten(self, indent: int = None) -> str:
        """
        将对象序列化为标准 cq action json 字符串，一般供连接器使用
        """
        return json.dumps(self.extract(), ensure_ascii=False, indent=indent)


class MsgPacker(ActionPacker):
    """
    消息 action 信息构造类
    """

    def __init__(
        self,
        msgs: List[CQMsgDict],
        isPrivate: bool,
        userId: int = None,
        groupId: int = None,
    ) -> None:
        super().__init__()
        self.type = "send_msg"
        if isPrivate:
            self.params = {
                "message_type": "private",
                "user_id": userId,
                "message": msgs,
            }
        else:
            self.params = {
                "message_type": "group",
                "user_id": userId,
                "group_id": groupId,
                "message": msgs,
            }


def msg_action(
    content: Union[str, CQMsgDict, List[CQMsgDict]],
    isPrivate: bool,
    userId: int = None,
    groupId: int = None,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    发送消息 action 构造方法
    """
    if isinstance(content, str):
        msgs = text_msg(content)
        if not isinstance(msgs, list):
            msgs = [msgs]
    elif isinstance(content, dict):
        msgs = [content]
    elif isinstance(content, list):
        temp = []
        for _ in content:
            if isinstance(_, list):
                temp.extend(_)
            else:
                temp.append(_)
        msgs = temp
    else:
        raise BotActionError("content 参数类型不正确，无法封装")
    return BotAction(
        MsgPacker(msgs, isPrivate, userId, groupId),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class ForwardMsgPacker(ActionPacker):
    """
    转发消息 action 信息构造类
    """

    def __init__(
        self,
        msgs: Dict,
        isPrivate: bool,
        userId: int = None,
        groupId: int = None,
    ) -> None:
        super().__init__()
        if isPrivate:
            self.type = "send_private_forward_msg"
            self.params = {"user_id": userId, "messages": msgs, "auto_escape": True}
        else:
            self.type = "send_group_forward_msg"
            self.params = {"group_id": groupId, "messages": msgs, "auto_escape": True}


def forward_msg_action(
    msgNodes: List[MsgNodeDict],
    isPrivate: bool,
    userId: int = None,
    groupId: int = None,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    转发消息发送 action 构造方法
    """
    return BotAction(
        ForwardMsgPacker(msgNodes, isPrivate, userId, groupId),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class MsgDelPacker(ActionPacker):
    """
    撤回消息 action 信息构造类
    """

    def __init__(self, msgId: int) -> None:
        super().__init__()
        self.type = "delete_msg"
        self.params = {
            "message_id": msgId,
        }


def msg_del_action(
    msgId: int, respWaited: bool = False, triggerEvent: BotEvent = None
) -> BotAction:
    """
    撤回消息 action 构造方法
    """
    return BotAction(
        MsgDelPacker(msgId), respWaited=respWaited, triggerEvent=triggerEvent
    )


class GetMsgPacker(ActionPacker):
    """
    消息信息获取 action 信息构造类
    """

    def __init__(self, msgId: int) -> None:
        super().__init__()
        self.type = "get_msg"
        self.params = {"message_id": msgId}


def get_msg_action(
    msgId: int, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取消息详细信息 action 构造方法
    """
    return BotAction(
        GetMsgPacker(msgId), respWaited=respWaited, triggerEvent=triggerEvent
    )


class getForwardPacker(ActionPacker):
    """
    转发消息获取 action 信息构造类
    """

    def __init__(self, forwardId: str) -> None:
        super().__init__()
        self.type = "get_forward_msg"
        self.params = {"message_id": forwardId}


def get_forward_msg_action(
    forwardId: str, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    转发消息获取 action 构造方法
    """
    return BotAction(
        getForwardPacker(forwardId), respWaited=respWaited, triggerEvent=triggerEvent
    )


class getImagePacker(ActionPacker):
    """
    获取图片信息 action 信息构造类
    """

    def __init__(self, fileName: str) -> None:
        super().__init__()
        self.type = "get_image"
        self.params = {"file": fileName}


def get_image_action(
    fileName: str, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取图片信息 action 构造方法
    """
    return BotAction(
        getImagePacker(fileName), respWaited=respWaited, triggerEvent=triggerEvent
    )


class MarkMsgReadPacker(ActionPacker):
    """
    标记消息已读 action 信息构造类
    """

    def __init__(self, msgId: int) -> None:
        super().__init__()
        self.type = "mark_msg_as_read"
        self.params = {"message_id": msgId}


def mark_msg_read_action(
    msgId: int, respWaited: bool = False, triggerEvent: BotEvent = None
) -> BotAction:
    """
    标记消息已读 action 构造方法
    """
    return BotAction(
        MarkMsgReadPacker(msgId), respWaited=respWaited, triggerEvent=triggerEvent
    )


class GroupKickPacker(ActionPacker):
    """
    群组踢人 action 信息构造类
    """

    def __init__(self, groupId: int, userId: int, laterReject: bool = False) -> None:
        super().__init__()
        self.type = "set_group_kick"
        self.params = {
            "group_id": groupId,
            "user_id": userId,
            "reject_add_request": laterReject,
        }


def group_kick_action(
    groupId: int,
    userId: int,
    laterReject: bool = False,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    群组踢人 action 构造方法
    """
    return BotAction(
        GroupKickPacker(groupId, userId, laterReject),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GroupBanPacker(ActionPacker):
    """
    群组禁言 action 信息构造类
    """

    def __init__(self, groupId: int, userId: int, duration: int) -> None:
        super().__init__()
        self.type = "set_group_ban"
        self.params = {
            "group_id": groupId,
            "user_id": userId,
            "duration": duration,
        }


def group_ban_action(
    groupId: int,
    userId: int,
    duration: int,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    群组禁言 action 构造方法。
    duration 为 0 取消禁言
    """
    return BotAction(
        GroupBanPacker(groupId, userId, duration),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GroupAnonymBanPacker(ActionPacker):
    """
    群组匿名禁言 action 信息构造类
    """

    def __init__(self, groupId: int, anonymFlag: str, duration: int) -> None:
        super().__init__()
        self.type = "set_group_anonymous_ban"
        self.params = {
            "group_id": groupId,
            "anonymous_flag": anonymFlag,
            "duration": duration,
        }


def group_anonym_ban_action(
    groupId: int,
    anonymFlag: str,
    duration: int,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    群组匿名禁言 action 构造方法。
    无法取消禁言
    """
    return BotAction(
        GroupAnonymBanPacker(groupId, anonymFlag, duration),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GroupWholeBanPacker(ActionPacker):
    """
    群组全员禁言 action 信息构造类
    """

    def __init__(self, groupId: int, enable: bool) -> None:
        super().__init__()
        self.type = "set_group_whole_ban"
        self.params = {"group_id": groupId, "enable": enable}


def group_whole_ban_action(
    groupId: int, enable: bool, respWaited: bool = False, triggerEvent: BotEvent = None
) -> BotAction:
    """
    群组全员禁言 action 构造方法
    """
    return BotAction(
        GroupWholeBanPacker(groupId, enable),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class SetGroupAdminPacker(ActionPacker):
    """
    设置群管理员 action 信息构造类
    """

    def __init__(self, groupId: int, userId: int, enable: bool) -> None:
        super().__init__()
        self.type = "set_group_admin"
        self.params = {"group_id": groupId, "user_id": userId, "enable": enable}


def set_group_admin_action(
    groupId: int,
    userId: int,
    enable: bool,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    设置群管理员 action 构造方法
    """
    return BotAction(
        SetGroupAdminPacker(groupId, userId, enable),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class SetGroupCardPacker(ActionPacker):
    """
    设置群名片 action 信息构造类
    """

    def __init__(self, groupId: int, userId: int, card: str) -> None:
        super().__init__()
        self.type = "set_group_card"
        self.params = {"group_id": groupId, "user_id": userId, "card": card}


def set_group_card_action(
    groupId: int,
    userId: int,
    card: str,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    设置群名片 action 构造方法
    """
    return BotAction(
        SetGroupCardPacker(groupId, userId, card),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class SetGroupNamePacker(ActionPacker):
    """
    设置群名 action 信息构造类
    """

    def __init__(self, groupId: int, name: str) -> None:
        super().__init__()
        self.type = "set_group_name"
        self.params = {"group_id": groupId, "group_name": name}


def set_group_name_action(
    groupId: int, name: str, respWaited: bool = False, triggerEvent: BotEvent = None
) -> BotAction:
    """
    设置群名 action 信息构造方法
    """
    return BotAction(
        SetGroupNamePacker(groupId, name),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GroupLeavePacker(ActionPacker):
    """
    退出群组 action 信息构造类
    """

    def __init__(self, groupId: int, isDismiss: bool) -> None:
        super().__init__()
        self.type = "set_group_leave"
        self.params = {"group_id": groupId, "is_dismiss": isDismiss}


def group_leave_action(
    groupId: int,
    isDismiss: bool,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    退出群组 action 构造方法
    """
    return BotAction(
        GroupLeavePacker(groupId, isDismiss),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class SetGroupTitlePacker(ActionPacker):
    """
    设置群头衔 action 信息构造类
    """

    def __init__(
        self,
        groupId: int,
        userId: int,
        title: str,
        duration: int = -1,
    ) -> None:
        super().__init__()
        self.type = "set_group_special_title"
        self.params = {
            "group_id": groupId,
            "user_id": userId,
            "special_title": title,
            "duration": duration,
        }


def set_group_title_action(
    groupId: int,
    userId: int,
    title: str,
    duration: int = -1,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    设置群头衔 action 构造方法
    """
    return BotAction(
        SetGroupTitlePacker(groupId, userId, title, duration),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GroupSignPacker(ActionPacker):
    """
    群打卡 action 信息构造类
    """

    def __init__(self, groupId: int) -> None:
        super().__init__()
        self.type = "send_group_sign"
        self.params = {"group_id": groupId}


def group_sign_action(
    groupId: int, respWaited: bool = False, triggerEvent: BotEvent = None
) -> BotAction:
    """
    群打卡 action 构造方法
    """
    return BotAction(
        GroupSignPacker(groupId), respWaited=respWaited, triggerEvent=triggerEvent
    )


class SetFriendAddPacker(ActionPacker):
    def __init__(self, addFlag: str, approve: bool, remark: str) -> None:
        """
        处理加好友请求 action 信息构造类
        """
        super().__init__()
        self.type = "set_friend_add_request"
        self.params = {"flag": addFlag, "approve": approve, "remark": remark}


def set_friend_add_action(
    addFlag: str,
    approve: bool,
    remark: str,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    处理加好友信息 action 构造方法。注意 remark 目前暂未实现
    """
    return BotAction(
        SetFriendAddPacker(addFlag, approve, remark),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class SetGroupAddPacker(ActionPacker):
    """
    处理加群请求 action 信息构造类
    """

    def __init__(
        self,
        addFlag: str,
        addType: Literal["add", "invite"],
        approve: bool,
        reason: str = None,
    ) -> None:
        super().__init__()
        self.type = "set_group_add_request"
        self.params = {
            "flag": addFlag,
            "sub_type": addType,
            "approve": approve,
        }
        if reason:
            self.params["reason"] = reason


def set_group_add_action(
    addFlag: str,
    addType: Literal["add", "invite"],
    approve: bool,
    rejectReason: str = None,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    处理加群请求 action 构造方法
    """
    return BotAction(
        SetGroupAddPacker(addFlag, addType, approve, rejectReason),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetLoginInfoPacker(ActionPacker):
    """
    获取登录号信息 action 信息构造类
    """

    def __init__(self) -> None:
        super().__init__()
        self.type = "get_login_info"
        self.params = {}


def get_login_info_action(
    respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获得登录号信息 action 构造方法
    """
    return BotAction(
        GetLoginInfoPacker(), respWaited=respWaited, triggerEvent=triggerEvent
    )


class SetLoginProfilePacker(ActionPacker):
    """
    设置登录号资料 action 信息构造类
    """

    def __init__(
        self, nickname: str, company: str, email: str, college: str, personalNote: str
    ) -> None:
        super().__init__()
        self.type = "set_qq_profile"
        self.params = {
            "nickname": nickname,
            "company": company,
            "email": email,
            "college": college,
            "personal_note": personalNote,
        }


def set_login_profile_action(
    nickname: str,
    company: str,
    email: str,
    college: str,
    personalNote: str,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    设置登录号资料 action 构造方法
    """
    return BotAction(
        SetLoginProfilePacker(nickname, company, email, college, personalNote),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetStrangerInfoPacker(ActionPacker):
    """
    获取陌生人信息 action 信息构造类
    """

    def __init__(self, userId: int, noCache: bool) -> None:
        super().__init__()
        self.type = "get_stranger_info"
        self.params = {"user_id": userId, "no_cache": noCache}


def get_stranger_info_action(
    userId: int, noCache: bool, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取陌生人信息 action 构造方法。也可以对好友使用
    """
    return BotAction(
        GetStrangerInfoPacker(userId, noCache),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetFriendListPacker(ActionPacker):
    """
    获取好友列表 action 信息构造类
    """

    def __init__(self) -> None:
        super().__init__()
        self.type = "get_friend_list"
        self.params = {}


def get_friend_list_action(
    respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取好友列表 action 构造方法
    """
    return BotAction(
        GetFriendListPacker(), respWaited=respWaited, triggerEvent=triggerEvent
    )


class GetUndirectFriendPacker(ActionPacker):
    """
    获取单向好友列表 action 信息构造类
    """

    def __init__(self) -> None:
        super().__init__()
        self.type = "get_unidirectional_friend_list"
        self.params = {}


def get_undirect_friend_action(
    respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取单向好友信息列表 action 构造方法
    """
    return BotAction(
        GetUndirectFriendPacker(), respWaited=respWaited, triggerEvent=triggerEvent
    )


class DeleteFriendPacker(ActionPacker):
    """
    删除好友 action 信息构造类
    """

    def __init__(self, userId: int) -> None:
        super().__init__()
        self.type = "delete_friend"
        self.params = {"user_id": userId}


def delete_friend_action(
    userId: int, respWaited: bool = False, triggerEvent: BotEvent = None
) -> BotAction:
    """
    删除好友 action 构造方法
    """
    return BotAction(
        DeleteFriendPacker(userId), respWaited=respWaited, triggerEvent=triggerEvent
    )


class GetGroupInfoPacker(ActionPacker):
    """
    获取群信息 action 信息构造类
    """

    def __init__(self, groupId: int, noCache: bool) -> None:
        super().__init__()
        self.type = "get_group_info"
        self.params = {"group_id": groupId, "no_cache": noCache}


def get_group_info_action(
    groupId: int, noCache: bool, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取群信息 action 构造方法。可以是未加入的群聊
    """
    return BotAction(
        GetGroupInfoPacker(groupId, noCache),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetGroupListPacker(ActionPacker):
    """
    获取群列表 action 信息构造类
    """

    def __init__(self) -> None:
        super().__init__()
        self.type = "get_group_list"
        self.params = {}


def get_group_list_action(
    respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取群列表 action 构造方法。注意返回建群时间都是 0，这是不准确的。准确的建群时间可以通过 `get_group_info_action` 获得
    """
    return BotAction(
        GetGroupListPacker(), respWaited=respWaited, triggerEvent=triggerEvent
    )


class GetGroupMemberInfoPacker(ActionPacker):
    """
    获取群成员信息 action 信息构造类
    """

    def __init__(self, groupId: int, userId: int, noCache: bool) -> None:
        super().__init__()
        self.type = "get_group_member_info"
        self.params = {"group_id": groupId, "user_id": userId, "no_cache": noCache}


def get_group_member_info_action(
    groupId: int,
    userId: int,
    noCache: bool,
    respWaited: bool = True,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    获取群成员信息 action 构造方法
    """
    return BotAction(
        GetGroupMemberInfoPacker(groupId, userId, noCache),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetGroupMemberListPacker(ActionPacker):
    """
    获取群成员列表 action 信息构造类
    """

    def __init__(self, groupId: int, noCache: bool) -> None:
        super().__init__()
        self.type = "get_group_member_list"
        self.params = {"group_id": groupId, "no_cache": noCache}


def get_group_member_list_action(
    groupId: int, noCache: bool, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取群成员列表 action 构造方法
    """
    return BotAction(
        GetGroupMemberListPacker(groupId, noCache),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetGroupHonorPacker(ActionPacker):
    """
    获取群荣誉信息 action 信息构造类
    """

    def __init__(
        self,
        groupId: int,
        type: Literal[
            "talkative", "performer", "legend", "strong_newbie", "emotion", "all"
        ],
    ) -> None:
        super().__init__()
        self.type = "get_group_honor_info"
        self.params = {"group_id": groupId, "type": type}


def get_group_honor_action(
    groupId: int,
    type: Literal[
        "talkative", "performer", "legend", "strong_newbie", "emotion", "all"
    ],
    respWaited: bool = True,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    获取群荣誉信息 action 构造方法
    """
    return BotAction(
        GetGroupHonorPacker(groupId, type),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class CheckSendImagePacker(ActionPacker):
    """
    检查是否可以发送图片 action 信息构造类
    """

    def __init__(self) -> None:
        super().__init__()
        self.type = "can_send_image"
        self.params = {}


def check_send_image_action(
    respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    检查是否可以发送图片 action 构造方法
    """
    return BotAction(
        CheckSendImagePacker(), respWaited=respWaited, triggerEvent=triggerEvent
    )


class CheckSendRecordPacker(ActionPacker):
    """
    检查是否可以发送语音 action 信息构造类
    """

    def __init__(self) -> None:
        super().__init__()
        self.type = "can_send_image"
        self.params = {}


def check_send_record_action(
    respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    检查是否可以发送语音 action 构造方法
    """
    return BotAction(
        CheckSendRecordPacker(), respWaited=respWaited, triggerEvent=triggerEvent
    )


class GetCqVersionPacker(ActionPacker):
    """
    获取 go-cqhttp 版本信息 action 信息构造类
    """

    def __init__(self) -> None:
        super().__init__()
        self.type = "get_version_info"
        self.params = {}


def get_cq_version_action(
    respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取 go-cqhttp 版本信息 action 构造方法
    """
    return BotAction(
        GetCqVersionPacker(), respWaited=respWaited, triggerEvent=triggerEvent
    )


class SetGroupPortraitPacker(ActionPacker):
    """
    设置群头像 action 信息构造类
    """

    def __init__(self, groupId: int, file: str, cache: Literal[0, 1] = 0) -> None:
        super().__init__()
        self.type = "set_group_portrait"
        self.params = {"group_id": groupId, "file": file, "cache": cache}


def set_group_portrait_action(
    groupId: int,
    file: str,
    cache: Literal[0, 1] = 0,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    设置群头像 action 构造方法。file 参数接受本地或网络 url 和 base64 编码。
    如本地路径为：`file:///C:/Users/Richard/Pictures/1.png`。
    特别注意：目前此 API 在登录一段时间后会因 cookie 失效而失效
    """
    return BotAction(
        SetGroupPortraitPacker(groupId, file, cache),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class OcrPacker(ActionPacker):
    """
    图片 OCR action 信息构造类
    """

    def __init__(self, image: str) -> None:
        super().__init__()
        self.type = "ocr_image"
        self.params = {"image": image}


def ocr_action(
    image: str, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    图片 OCR action 构造方法。image 为图片 ID
    """
    return BotAction(OcrPacker(image), respWaited=respWaited, triggerEvent=triggerEvent)


class GetGroupSysMsgPacker(ActionPacker):
    """
    获取群系统消息 action 信息构造类
    """

    def __init__(self) -> None:
        super().__init__()
        self.type = "get_group_system_msg"
        self.params = {}


def get_group_sys_msg_action(
    respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取群系统消息 action 构造方法
    """
    return BotAction(
        GetGroupSysMsgPacker(), respWaited=respWaited, triggerEvent=triggerEvent
    )


class UploadFilePacker(ActionPacker):
    """
    发送文件 action 信息构造类
    """

    def __init__(
        self,
        isPrivate: bool,
        file: str,
        name: str,
        userId: int = None,
        groupId: int = None,
        groupFolderId: str = None,
    ) -> None:
        super().__init__()
        if isPrivate:
            self.type = "upload_private_file"
            self.params = {"user_id": userId, "file": file, "name": name}
        else:
            self.type = "upload_group_file"
            self.params = {
                "group_id": groupId,
                "file": file,
                "name": name,
                "folder": groupFolderId,
            }


def upload_file_action(
    isPrivate: bool,
    file: str,
    sendFileName: str,
    userId: int = None,
    groupId: int = None,
    groupFolderId: str = None,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    发送文件 action 构造方法。只支持发送本地文件。
    若为群聊文件发送，不提供 folder id，则默认上传到群文件根目录。

    示例路径：`C:/users/15742/desktop/QQ图片20230108225606.jpg`。

    （若需要发送网络文件，先使用 `download_file_action()` 方法生成下载网络文件的 action，
    action 响应后文件会放于 go-cqhttp 缓存文件夹中，可直接在消息段中引用）
    """
    return BotAction(
        UploadFilePacker(isPrivate, file, sendFileName, userId, groupId, groupFolderId),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetGroupFileSysInfoPacker(ActionPacker):
    """
    获取群文件系统信息 action 信息构造类
    """

    def __init__(self, groupId: int) -> None:
        super().__init__()
        self.type = "get_group_file_system_info"
        self.params = {"group_id": groupId}


def get_group_filesys_info_action(
    groupId: int, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取群文件系统信息 action 构造方法
    """
    return BotAction(
        GetGroupFileSysInfoPacker(groupId),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetGroupRootFilesPacker(ActionPacker):
    """
    获取群根目录文件列表 action 信息构造类
    """

    def __init__(self, groupId: int) -> None:
        super().__init__()
        self.type = "get_group_root_files"
        self.params = {"group_id": groupId}


def get_group_root_files_action(
    groupId: int, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取群根目录文件列表 action 构造方法
    """
    return BotAction(
        GetGroupRootFilesPacker(groupId),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetGroupFilesByFolderPacker(ActionPacker):
    """
    获取群子目录文件列表 action 信息构造类
    """

    def __init__(self, groupId: int, folderId: str) -> None:
        super().__init__()
        self.type = "get_group_files_by_folder"
        self.params = {"group_id": groupId, "folder_id": folderId}


def get_group_files_byfolder_action(
    groupId: int, folderId: str, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取群子目录文件列表 action 构造方法
    """
    return BotAction(
        GetGroupFilesByFolderPacker(groupId, folderId),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class CreateGroupFolderPacker(ActionPacker):
    """
    创建群文件夹 action 信息构造类
    """

    def __init__(self, groupId: int, folderName: str) -> None:
        super().__init__()
        self.type = "create_group_file_folder"
        self.params = {"group_id": groupId, "name": folderName, "parent_id": "/"}


def create_group_folder_action(
    groupId: int,
    folderName: str,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    创建群文件夹 action 构造方法。注意：只能在根目录创建文件夹
    """
    return BotAction(
        CreateGroupFolderPacker(groupId, folderName),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class DeleteGroupFolderPacker(ActionPacker):
    """
    删除群文件夹 action 信息构造类
    """

    def __init__(self, groupId: int, folderId: str) -> None:
        super().__init__()
        self.type = "delete_group_folder"
        self.params = {"group_id": groupId, "folder_id": folderId}


def delete_group_folder_action(
    groupId: int, folderId: str, respWaited: bool = False, triggerEvent: BotEvent = None
) -> BotAction:
    """
    删除群文件夹 action 构造方法。
    """
    return BotAction(
        DeleteGroupFolderPacker(groupId, folderId),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class DeleteGroupFilePacker(ActionPacker):
    """
    删除群文件 action 信息构造类
    """

    def __init__(self, groupId: int, fileId: str, fileTypeId: int) -> None:
        super().__init__()
        self.type = "delete_group_file"
        self.params = {"group_id": groupId, "file_id": fileId, "busid": fileTypeId}


def delete_group_file_action(
    groupId: int,
    fileId: str,
    fileTypeId: int,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    删除群文件 action 构造方法。文件相关信息通过 `get_group_root_files_action()` 或
    `get_group_files_action` 的响应获得
    """
    return BotAction(
        DeleteGroupFilePacker(groupId, fileId, fileTypeId),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetGroupFileUrlPacker(ActionPacker):
    """
    获取群文件资源链接 action 信息构造类
    """

    def __init__(self, groupId: int, fileId: str, fileTypeId: int) -> None:
        super().__init__()
        self.type = "get_group_file_url"
        self.params = {"group_id": groupId, "file_id": fileId, "busid": fileTypeId}


def get_group_file_url_action(
    groupId: int,
    fileId: str,
    fileTypeId: int,
    respWaited: bool = True,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    获取群文件资源链接 action 构造方法。文件相关信息通过 `get_group_root_files_action()` 或
    `get_group_files_action` 的响应获得
    """
    return BotAction(
        GetGroupFileUrlPacker(groupId, fileId, fileTypeId),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetCqStatusPacker(ActionPacker):
    """
    获取 go-cqhttp 状态 action 信息构造类
    """

    def __init__(self) -> None:
        super().__init__()
        self.type = "get_status"
        self.params = {}


def get_cq_status_action(
    respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取 go-cqhttp 状态 action 构造方法
    """
    return BotAction(
        GetCqStatusPacker(), respWaited=respWaited, triggerEvent=triggerEvent
    )


class GetAtAllRemainPacker(ActionPacker):
    """
    获取群 @全体成员 剩余次数 action 信息构造类
    """

    def __init__(self, groupId: int) -> None:
        super().__init__()
        self.type = "get_group_at_all_remain"
        self.params = {"group_id": groupId}


def get_atall_remain_action(
    groupId: int, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取群 @全体成员 剩余次数 action 构造方法
    """
    return BotAction(
        GetAtAllRemainPacker(groupId), respWaited=respWaited, triggerEvent=triggerEvent
    )


class QuickHandlePacker(ActionPacker):
    """
    事件快速操作 action 信息构造类
    """

    def __init__(self, contextEvent: BotEvent, operation: dict) -> None:
        super().__init__()
        self.type = ".handle_quick_operation"
        self.params = {"context": contextEvent.raw, "operation": operation}


def quick_handle_action(
    contextEvent: BotEvent,
    operation: dict,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    事件快速操作 action 构造方法
    """
    return BotAction(
        QuickHandlePacker(contextEvent, operation),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class SetGroupNoticePacker(ActionPacker):
    """
    发送群公告 action 信息构造类
    """

    def __init__(self, groupId: int, content: str, imageUrl: str = None) -> None:
        super().__init__()
        self.type = "_send_group_notice"
        self.params = {
            "group_id": groupId,
            "content": content,
        }
        if imageUrl:
            self.params["image"] = imageUrl


def set_group_notice_action(
    groupId: int,
    content: str,
    imageUrl: str = None,
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    发送群公告 action 构造方法。注意 `imageUrl` 只能为本地 url，示例：`file:///C:/users/15742/desktop/123.jpg`
    """
    return BotAction(
        SetGroupNoticePacker(groupId, content, imageUrl),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetGroupNoticePacker(ActionPacker):
    """
    获取群公告 action 信息构造类
    """

    def __init__(
        self,
        groupId: int,
    ) -> None:
        super().__init__()
        self.type = "_get_group_notice"
        self.params = {
            "group_id": groupId,
        }


def get_group_notice_action(
    groupId: int, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取群公告 action 构造方法。
    群公告图片有 id，但暂时没有下载的方法
    """
    return BotAction(
        GetGroupNoticePacker(groupId), respWaited=respWaited, triggerEvent=triggerEvent
    )


class DownloadFilePacker(ActionPacker):
    """
    下载文件到缓存目录 action 信息构造类
    """

    def __init__(self, fileUrl: str, useThreadNum: int, headers: dict) -> None:
        super().__init__()
        self.type = "download_file"
        self.params = {"url": fileUrl, "thread_count": useThreadNum, "headers": headers}


def download_file_action(
    fileUrl: str,
    useThreadNum: int,
    headers: Union[List, str],
    respWaited: bool = True,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    下载文件到缓存目录 action 构造方法。`headers` 的两种格式：
    ```
    "User-Agent=YOUR_UA[\\r\\n]Referer=https://www.baidu.com"
    ```
    或
    ```python
    [
        "User-Agent=YOUR_UA",
        "Referer=https://www.baidu.com"
    ]
    ```
    """
    return BotAction(
        DownloadFilePacker(fileUrl, useThreadNum, headers),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetOnlineClientsPacker(ActionPacker):
    """
    获取当前账号在线客户端列表 action 信息构造类
    """

    def __init__(self, noCache: bool) -> None:
        super().__init__()
        self.type = "get_online_clients"
        self.params = {"no_cache": noCache}


def get_online_clients_action(
    noCache: bool, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取当前账号在线客户端列表 action 构造方法
    """
    return BotAction(
        GetOnlineClientsPacker(noCache),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetGroupMsgHistoryPacker(ActionPacker):
    """
    获取群消息历史记录 action 信息构造类
    """

    def __init__(self, msgSeq: int, groupId: int) -> None:
        super().__init__()
        self.type = "get_group_msg_history"
        self.params = {"message_seq": msgSeq, "group_id": groupId}


def get_group_msg_history_action(
    msgSeq: int, groupId: int, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取群消息历史记录 action 构造方法
    """
    return BotAction(
        GetGroupMsgHistoryPacker(msgSeq, groupId),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class SetGroupEssencePacker(ActionPacker):
    """
    设置精华消息 action 信息构造类
    """

    def __init__(self, msgId: int, type: Literal["add", "del"]) -> None:
        super().__init__()
        if type == "add":
            self.type = "set_essence_msg"
        else:
            self.type = "delete_essence_msg"
        self.params = {"message_id": msgId}


def set_group_essence_action(
    msgId: int,
    type: Literal["add", "del"],
    respWaited: bool = False,
    triggerEvent: BotEvent = None,
) -> BotAction:
    """
    设置精华消息 action 构造方法
    """
    return BotAction(
        SetGroupEssencePacker(msgId, type),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetGroupEssenceListPacker(ActionPacker):
    """
    获取精华消息列表 action 信息构造类
    """

    def __init__(self, groupId: int) -> None:
        super().__init__()
        self.type = "get_essence_msg_list"
        self.params = {"group_id": groupId}


def get_group_essence_list_action(
    groupId: int, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取精华消息列表 action 构造方法
    """
    return BotAction(
        GetGroupEssenceListPacker(groupId),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class GetModelShowPacker(ActionPacker):
    """
    获取在线机型 action 信息构造类
    """

    def __init__(self, model: str) -> None:
        super().__init__()
        self.type = "_get_model_show"
        self.params = {"model": model}


def get_model_show_action(
    model: str, respWaited: bool = True, triggerEvent: BotEvent = None
) -> BotAction:
    """
    获取在线机型 action 构造方法
    """
    return BotAction(
        GetModelShowPacker(model), respWaited=respWaited, triggerEvent=triggerEvent
    )


class SetModelShowPacker(ActionPacker):
    """
    设置在线机型 action 信息构造类
    """

    def __init__(self, model: str, modelShow: str) -> None:
        super().__init__()
        self.type = "_set_model_show"
        self.params = {"model": model, "model_show": modelShow}


def set_model_show_action(
    model: str, modelShow: str, respWaited: bool = False, triggerEvent: BotEvent = None
) -> BotAction:
    """
    设置在线机型 action 构造方法
    """
    return BotAction(
        SetModelShowPacker(model, modelShow),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )


class DeleteUndirectFriendPacker(ActionPacker):
    """
    删除单向好友 action 信息构造类
    """

    def __init__(self, userId: int) -> None:
        super().__init__()
        self.type = "delete_unidirectional_friend"
        self.params = {"user_id": userId}


def delete_undirect_friend_action(
    userId: int, respWaited: bool = False, triggerEvent: BotEvent = None
) -> BotAction:
    """
    删除单向好友 action 构造方法
    """
    return BotAction(
        DeleteUndirectFriendPacker(userId),
        respWaited=respWaited,
        triggerEvent=triggerEvent,
    )



import re

from ..types.typing import *
from ..types.utils import BotMatcher


class AlwaysMatch(BotMatcher):
    def __init__(self) -> None:
        super().__init__()

    def match(self, text: str) -> bool:
        return True


class StartMatch(BotMatcher):
    def __init__(self, target: str) -> None:
        super().__init__()
        self.target = target

    def match(self, text: str) -> bool:
        return text[: len(self.target)] == self.target


class ContainMatch(BotMatcher):
    def __init__(self, target: str, freq: int = 1) -> None:
        super().__init__()
        self.target = target
        self.freq = freq

    def match(self, text: str) -> bool:
        if self.freq == 1:
            return self.target in text
        else:
            return len(re.findall(self.target, text)) == self.freq


class EndMatch(BotMatcher):
    def __init__(self, target: str) -> None:
        super().__init__()
        self.target = target

    def match(self, text: str) -> bool:
        return text[-len(self.target) :] == self.target


class FullMatch(BotMatcher):
    def __init__(self, target: str) -> None:
        super().__init__()
        self.target = target

    def match(self, text: str) -> bool:
        return text == self.target


class RegexMatch(BotMatcher):
    def __init__(self, regex_pattern: str, regex_flags: Any = 0) -> None:
        super().__init__()
        self.pattern = regex_pattern
        self.flag = regex_flags

    def match(self, text: str) -> bool:
        return len(re.findall(self.pattern, text, self.flag)) > 0



import asyncio as aio
import traceback

from ..models.bot import BotHookBus
from ..models.event import BotEvent
from ..models.handler import (
    EventHandler,
    MetaEventHandler,
    MsgEventHandler,
    NoticeEventHandler,
    ReqEventHandler,
)
from ..types.core import IEventDispatcher
from ..types.exceptions import *
from ..types.models import BotLife
from ..types.typing import *
from ..utils.logger import Logger


class BotDispatcher(IEventDispatcher):
    """
    bot 调度模块。负责将传递的普通事件送入各事件总线
    （接收的事件类型：消息、请求和通知）
    """

    def __init__(self, logger: Logger) -> None:
        super().__init__()
        self._handlers: Dict[str, List[EventHandler]] = {
            "message": [],
            "request": [],
            "notice": [],
            "meta": [],
        }
        self.logger = logger

        self._ready_signal = aio.Event()

    def add_handlers(self, all_handlers: List[EventHandler]) -> None:
        """
        绑定事件处理器列表
        """
        self._ready_signal.clear()

        for handler in all_handlers:
            if isinstance(handler, MsgEventHandler):
                self._handlers["message"].append(handler)
            elif isinstance(handler, ReqEventHandler):
                self._handlers["request"].append(handler)
            elif isinstance(handler, NoticeEventHandler):
                self._handlers["notice"].append(handler)
            elif isinstance(handler, MetaEventHandler):
                self._handlers["meta"].append(handler)
        for k in self._handlers.keys():
            self._handlers[k] = sorted(
                self._handlers[k], key=lambda x: x.priority, reverse=True
            )

        self._ready_signal.set()

    async def dispatch(self, event: BotEvent) -> None:
        """
        把事件分发到对应的事件总线
        """
        await self._ready_signal.wait()
        await BotHookBus.emit(BotLife.EVENT_RECEIVED, event, wait=True)

        try:
            permit_priority = PriorityLevel.MIN.value
            handlers = self._handlers[event.type]
            for handler in handlers:
                # 事件处理器优先级不够，则不分配给它处理
                if handler.priority < permit_priority:
                    continue
                # evoke 返回的值用于判断，事件处理器内部经过各种检查后，是否选择处理这个事件。
                if not (await handler.evoke(event)):
                    # 如果决定不处理，则会跳过此次循环（也就是不进行“可能存在的优先级阻断操作”）
                    continue
                if handler.set_block and handler.priority > permit_priority:
                    permit_priority = handler.priority
        except Exception as e:
            self.logger.error(f"bot dispatcher 抛出异常：[{e.__class__.__name__}] {e}")
            self.logger.debug(f"异常点的事件记录为：{event.raw}")
            self.logger.debug("异常回溯栈：\n" + traceback.format_exc().strip("\n"))
